"""
    FilterPanel contains classes the implement the tracking of filters for the data panel.
"""

# standard libraries
import bisect
import datetime
import gettext
import threading
import weakref

# third party libraries
# None

# local libraries
from nion.swift import Panel
from nion.utils import ListModel

_ = gettext.gettext


# TODO: Add button to convert between (filter <-> smart group) -> regular group
# TODO: Add tree browser for sessions, organized by date
# TODO: Add tree browser for date
# TODO: Add list browser for users
# TODO: Add list browser for devices
# TODO: Add checkbox browser for flagged / not flagged
# TODO: Add star browser for rating
# TODO: Add tree browser for keywords
# TODO: Add text field browser for searching


class FilterController:

    """
        The FilterController creates, updates, and provides access to an item model controller.

        An item model controller is the controller associated with the UI system for displaying trees.

        The item model controller represents the years, months, and days available within the list of
        data items.
    """

    def __init__(self, document_controller):
        self.ui = document_controller.ui
        self.__periodic_listener = document_controller.add_periodic(1.0, self.__periodic)
        self.item_model_controller = self.ui.create_item_model_controller(["display"])
        self.__document_controller_weakref = weakref.ref(document_controller)

        self.__data_item_tree = TreeNode(reversed=True)
        self.__data_item_tree_mutex = threading.RLock()

        self.__data_item_tree.child_inserted = self.__insert_child
        self.__data_item_tree.child_removed = self.__remove_child
        self.__data_item_tree.tree_node_updated = self.__update_tree_node

        # thread safe.
        def data_item_inserted(key, data_item, before_index):
            """
                This method will be called from the data item list model, which comes from the document controller to
                notify that the list of data items in the document has changed.
                This method breaks the date-related metadata out into a list of indexes which are then displayed
                in tree format for the date browser. in this case, the indexes are added.
            """
            assert threading.current_thread() == threading.main_thread()
            with self.__data_item_tree_mutex:
                created_local = data_item.created_local
                indexes = created_local.year, created_local.month, created_local.day
                self.__data_item_tree.insert_value(indexes, data_item)

        # thread safe.
        def data_item_removed(key, data_item, index):
            """
                This method will be called from the data item list model, which comes from the document controller to
                notify that the list of data items in the document has changed.
                This method breaks the date-related metadata out into a list of indexes which are then displayed
                in tree format for the date browser. in this case, the indexes are removed.
            """
            assert threading.current_thread() == threading.main_thread()
            with self.__data_item_tree_mutex:
                created = data_item.created_local
                indexes = created.year, created.month, created.day
                self.__data_item_tree.remove_value(indexes, data_item)

        # connect the data_items_model from the document controller to self.
        # when data items are inserted or removed from the document controller, the inserter and remover methods
        # will be called.
        self.__data_items_model = document_controller.data_items_model

        self.__library_item_inserted_listener = self.__data_items_model.item_inserted_event.listen(data_item_inserted)
        self.__library_item_removed_listener = self.__data_items_model.item_removed_event.listen(data_item_removed)

        self.__mapping = dict()
        self.__mapping[id(self.__data_item_tree)] = self.item_model_controller.root
        self.__node_counts_dirty = False

        self.__date_filter = None
        self.__text_filter = None

        for index, data_item in enumerate(self.__data_items_model.data_items):
            data_item_inserted("data_items", data_item, index)

    def close(self):
        # Close the data model controller. Un-listen to the data item list model and close the item model controller.
        self.__library_item_inserted_listener.close()
        self.__library_item_inserted_listener = None
        self.__library_item_removed_listener.close()
        self.__library_item_removed_listener = None
        self.__periodic_listener.close()
        self.__periodic_listener = None
        self.item_model_controller.close()
        self.item_model_controller = None

    @property
    def document_controller(self):
        return self.__document_controller_weakref()

    def __periodic(self):
        self.update_all_nodes()

    def __display_for_tree_node(self, tree_node):
        """ Return the text display for the given tree node. Based on number of keys associated with tree node. """
        keys = tree_node.keys
        if len(keys) == 1:
            return "{0} ({1})".format(tree_node.keys[-1], tree_node.count)
        elif len(keys) == 2:
            months = (_("January"), _("February"), _("March"), _("April"), _("May"), _("June"), _("July"), _("August"),
                      _("September"), _("October"), _("November"), _("December"))
            return "{0} ({1})".format(months[max(min(tree_node.keys[1]-1, 11), 0)], tree_node.count)
        else:
            weekdays = (_("Monday"), _("Tuesday"), _("Wednesday"), _("Thursday"), _("Friday"), _("Saturday"), _("Sunday"))
            date = datetime.date(tree_node.keys[0], tree_node.keys[1], tree_node.keys[2])
            return "{0} - {1} ({2})".format(tree_node.keys[2], weekdays[date.weekday()], tree_node.count)

    def __insert_child(self, parent_tree_node, index, tree_node):
        """
            Called from the root tree node when a new node is inserted into tree. This method creates properties
            to represent the node for display and inserts it into the item model controller.
        """
        # manage the item model
        parent_item = self.__mapping[id(parent_tree_node)]
        self.item_model_controller.begin_insert(index, index, parent_item.row, parent_item.id)
        properties = {
            "display": self.__display_for_tree_node(tree_node),
            "tree_node": tree_node  # used for removal and other lookup
        }
        item = self.item_model_controller.create_item(properties)
        parent_item.insert_child(index, item)
        self.__mapping[id(tree_node)] = item
        self.item_model_controller.end_insert()

    def __remove_child(self, parent_tree_node, index):
        """
            Called from the root tree node when a node is removed from the tree. This method removes it into the
            item model controller.
        """
        # get parent and item
        parent_item = self.__mapping[id(parent_tree_node)]
        # manage the item model
        self.item_model_controller.begin_remove(index, index, parent_item.row, parent_item.id)
        child_item = parent_item.children[index]
        parent_item.remove_child(child_item)
        self.__mapping.pop(id(child_item.data["tree_node"]))
        self.item_model_controller.end_remove()

    def __update_tree_node(self, tree_node):
        """ Mark the fact that tree node counts need updating when convenient. """
        self.__node_counts_dirty = True

    def update_all_nodes(self):
        """ Update all tree item displays if needed. Usually for count updates. """
        item_model_controller = self.item_model_controller
        if item_model_controller:
            if self.__node_counts_dirty:
                for item in self.__mapping.values():
                    if "tree_node" in item.data:  # don't update the root node
                        tree_node = item.data["tree_node"]
                        item.data["display"] = self.__display_for_tree_node(tree_node)
                        item_model_controller.data_changed(item.row, item.parent.row, item.parent.id)
                self.__node_counts_dirty = False

    def date_browser_selection_changed(self, selected_indexes):
        """
            Called to handle selection changes in the tree widget.

            This method should be connected to the on_selection_changed event. This method builds a list
            of keys represented by all selected items. It then provides date_filter to filter data items
            based on the list of keys. It then sets the filter into the document controller.

            :param selected_indexes: The selected indexes
            :type selected_indexes: list of ints
        """
        partial_date_filters = list()

        for index, parent_row, parent_id in selected_indexes:
            item_model_controller = self.item_model_controller
            tree_node = item_model_controller.item_value("tree_node", index, parent_id)
            partial_date_filters.append(ListModel.PartialDateFilter("created_local", *tree_node.keys))

        if len(partial_date_filters) > 0:
            self.__date_filter = ListModel.OrFilter(partial_date_filters)
        else:
            self.__date_filter = None

        self.__update_filter()

    def text_filter_changed(self, text):
        """
            Called to handle changes to the text filter.

            :param text: The text for the filter.
        """
        text = text.strip() if text else None

        if text is not None:
            self.__text_filter = ListModel.TextFilter("text_for_filter", text)
        else:
            self.__text_filter = None

        self.__update_filter()

    def __update_filter(self):
        """
            Create a combined filter. Set the resulting filter into the document controller.
        """
        filters = list()
        if self.__date_filter:
            filters.append(self.__date_filter)
        if self.__text_filter:
            filters.append(self.__text_filter)
        self.document_controller.display_filter = ListModel.AndFilter(filters)


class FilterPanel:

    """
        A object to hold the widget for the filter panel.
    """

    def __init__(self, document_controller):

        ui = document_controller.ui
        self.document_controller = document_controller

        self.__filter_controller = self.document_controller.filter_controller

        date_browser_tree_widget = ui.create_tree_widget()
        date_browser_tree_widget.selection_mode = "extended"
        date_browser_tree_widget.item_model_controller = self.__filter_controller.item_model_controller
        date_browser_tree_widget.on_selection_changed = self.__filter_controller.date_browser_selection_changed

        date_browser = ui.create_column_widget()
        date_browser.add(ui.create_label_widget(_("Date"), properties={"stylesheet": "font-weight: bold"}))
        date_browser.add(date_browser_tree_widget)
        date_browser.add_stretch()

        header_canvas_item = Panel.HeaderCanvasItem(document_controller, _("Filter"))

        header_widget = ui.create_canvas_widget(properties={"height": header_canvas_item.header_height})
        header_widget.canvas_item.add_canvas_item(header_canvas_item)

        filter_bar_row = ui.create_row_widget()
        filter_bar_row.add(ui.create_label_widget(_("Search")))
        filter_text_widget = ui.create_line_edit_widget(properties={"width": 160})
        filter_text_widget.placeholder_text = _("No Filter")
        filter_text_widget.on_text_edited = self.__filter_controller.text_filter_changed
        clear_filter_text_widget = ui.create_push_button_widget(_("Clear"))
        def clear_filter():
            filter_text_widget.text = ""
            self.__filter_controller.text_filter_changed("")
        clear_filter_text_widget.on_clicked = clear_filter
        filter_bar_row.add_spacing(8)
        filter_bar_row.add(filter_text_widget)
        filter_bar_row.add_spacing(8)
        filter_bar_row.add(clear_filter_text_widget)
        filter_bar_row.add_stretch()

        filter_column = ui.create_column_widget(properties={"height": 180})
        filter_column.add(header_widget)
        filter_column.add_spacing(4)
        filter_column.add(filter_bar_row)
        filter_column.add_spacing(4)
        filter_column.add(date_browser)
        filter_column.add_spacing(4)

        self.widget = filter_column


class TreeNode:

    """
        Represents a node in a tree, used for implementing data item filters.

        Tracks cumulative child count.
    """

    def __init__(self, key=None, children=None, values=None, reversed=False):
        self.key = key
        self.count = 0
        self.reversed = reversed
        self.__weak_parent = None
        self.children = children if children is not None else list()
        self.values = values if values is not None else list()
        self.__value_reverse_mapping = dict()
        self.child_inserted = None
        self.child_removed = None
        self.tree_node_updated = None

    def __lt__(self, other):
        return self.key < other.key if not self.reversed else other.key < self.key

    def __le__(self, other):
        return self.key <= other.key if not self.reversed else other.key <= self.key

    def __eq__(self, other):
        return self.key == other.key

    def __ne__(self, other):
        return self.key != other.key

    def __gt__(self, other):
        return self.key > other.key if not self.reversed else other.key > self.key

    def __ge__(self, other):
        return self.key >= other.key if not self.reversed else other.key >= self.key

    def __hash__(self):
        return self.key.__hash__()

    def __repr__(self):
        return "{0}/{1}:{2}{3}".format(self.key, self.count, " {0}".format(self.children) if self.children else str(),
                                       " <{0}>".format(len(self.values)) if self.values else str())

    def __get_parent(self):
        """
            Return the parent tree node, if any. Read only.

            :return: The parent tree node
            :rtype: :py:class:`nion.swift.FilterPanel.TreeNode`
        """
        return self.__weak_parent() if self.__weak_parent else None
    parent = property(__get_parent)

    def __set_parent(self, parent):
        """ Set the parent tree node. Private. """
        self.__weak_parent = weakref.ref(parent) if parent else None

    def __get_keys(self):
        """ Return the keys associated with this node by adding its key and then adding parent keys recursively. """
        keys = list()
        tree_node = self
        while tree_node is not None and tree_node.key is not None:
            keys.insert(0, tree_node.key)
            tree_node = tree_node.parent
        return keys
    keys = property(__get_keys)

    def insert_value(self, keys, value):
        """
            Insert a value (data item) into this tree node and then its
            children. This will be called in response to a new data item being
            inserted into the document. Also updates the tree node's cumulative
            child count.
        """
        self.count += 1
        if not self.key:
            self.__value_reverse_mapping[value] = keys
        if len(keys) == 0:
            self.values.append(value)
        else:
            key = keys[0]
            index = bisect.bisect_left(self.children, TreeNode(key, reversed=self.reversed))
            if index == len(self.children) or self.children[index].key != key:
                new_tree_node = TreeNode(key, list(), reversed=self.reversed)
                new_tree_node.child_inserted = self.child_inserted
                new_tree_node.child_removed = self.child_removed
                new_tree_node.tree_node_updated = self.tree_node_updated
                new_tree_node.__set_parent(self)
                self.children.insert(index, new_tree_node)
                if self.child_inserted:
                    self.child_inserted(self, index, new_tree_node)
            child = self.children[index]
            child.insert_value(keys[1:], value)
            if self.tree_node_updated:
                self.tree_node_updated(child)

    def remove_value(self, keys, value):
        """
            Remove a value (data item) from this tree node and its children.
            Also updates the tree node's cumulative child count.
        """
        self.count -= 1
        if not self.key:
            keys = self.__value_reverse_mapping[value]
            del self.__value_reverse_mapping[value]
        if len(keys) == 0:
            self.values.remove(value)
        else:
            key = keys[0]
            index = bisect.bisect_left(self.children, TreeNode(key, reversed=self.reversed))
            assert index != len(self.children) and self.children[index].key == key
            self.children[index].remove_value(keys[1:], value)
            if self.tree_node_updated:
                self.tree_node_updated(self.children[index])
            if self.children[index].count == 0:
                del self.children[index]
                if self.child_removed:
                    self.child_removed(self, index)



# standard libraries
import copy
import functools
import gettext
import json
import random
import string
import typing
import uuid
import weakref

# third party libraries
# None

# local libraries
from nion.swift import DisplayPanel
from nion.swift import Undo
from nion.swift.model import Utility
from nion.swift.model import WorkspaceLayout
from nion.ui import CanvasItem


_ = gettext.gettext


def create_image_desc():
    return {"type": "image", "identifier": "".join([random.choice(string.ascii_uppercase) for _ in range(2)]), "uuid": str(uuid.uuid4())}

def create_splitter_desc(orientation: str, splits, children):
    return {"type": "splitter", "orientation": orientation, "splits": copy.copy(splits), "children": children}


class Workspace:
    """
        The Workspace object keeps the overall layout of the window. It contains
        root item which contains tab groups arranged into boxes (rows and columns).
        The boxes contain other boxes or tab groups. Tab groups contain tabs which contain
        a single panel.

        The Workspace is able to save/restore itself from a text stream. It is also able to
        record changes to the workspace from the enclosing application. When restoring its
        layout, it is able to handle cases where the size/shape of the window is different
        from its original size/shape when saved.
    """

    def __init__(self, document_controller, workspace_id):
        self.__document_controller_weakref = weakref.ref(document_controller)

        self.ui = self.document_controller.ui

        self.workspace_manager = WorkspaceManager()

        self.workspace_id = workspace_id

        self.dock_widgets = []
        self.display_panels = []

        self.__canvas_item = None

        # create the root element
        root_widget = self.ui.create_column_widget(properties={"min-width": 640, "min-height": 480})
        self.__content_column = self.ui.create_column_widget()
        self.message_column = self.ui.create_column_widget()
        self.filter_panel = self.workspace_manager.create_filter_panel(document_controller)
        self.filter_row = self.filter_panel.widget
        self.image_row = self.ui.create_column_widget()
        self.__content_column.add(self.message_column)
        self.__content_column.add(self.filter_row)
        self.__content_column.add(self.image_row, fill=True)
        self.filter_row.visible = False
        root_widget.add(self.__content_column)

        self.__message_boxes = dict()

        # configure the document window (central widget)
        document_controller.attach_widget(root_widget)

        visible_panels = []
        if self.workspace_id == "library":
            visible_panels = ["toolbar-panel", "data-panel", "histogram-panel", "info-panel", "inspector-panel"]

        self.create_panels(visible_panels)

        self.__workspace = None

    def close(self):
        for message_box_widget in copy.copy(list(self.__message_boxes.values())):
            self.message_column.remove(message_box_widget)
        self.__message_boxes.clear()
        if self.__workspace:
            self.__sync_layout()
        self.display_panels = []
        self.__canvas_item = None
        self.__workspace = None
        dock_widgets_copy = copy.copy(self.dock_widgets)
        self.dock_widgets = None
        for dock_widget in dock_widgets_copy:
            dock_widget.panel.close()
            # dock_widget.close()  # closed by the panel
        self.__content_column = None
        self.filter_row = None
        self.image_row = None
        self.document_controller.detach_widget()

    def periodic(self):
        for dock_widget in self.dock_widgets if self.dock_widgets else list():
            dock_widget.panel.periodic()
            dock_widget.periodic()

    def restore_geometry_state(self):
        geometry = self.ui.get_persistent_string("Workspace/%s/Geometry" % self.workspace_id)
        state = self.ui.get_persistent_string("Workspace/%s/State" % self.workspace_id)
        return geometry, state

    def save_geometry_state(self, geometry, state):
        # ugh. this has the side effect of saving the layout when the geometry state is saved.
        self.__sync_layout()
        self.ui.set_persistent_string("Workspace/%s/Geometry" % self.workspace_id, geometry)
        self.ui.set_persistent_string("Workspace/%s/State" % self.workspace_id, state)

    @property
    def document_controller(self):
        return self.__document_controller_weakref()

    @property
    def document_model(self):
        return self.document_controller.document_model

    @property
    def _canvas_item(self):
        return self.__canvas_item

    def _find_dock_widget(self, dock_widget_id):
        for dock_widget in self.dock_widgets:
            if dock_widget.panel.panel_id == dock_widget_id:
                return dock_widget
        return None

    def create_panels(self, visible_panels=None):
        # get the document controller
        document_controller = self.document_controller

        # add registered panels
        for panel_id in self.workspace_manager.panel_ids:
            title, positions, position, properties = self.workspace_manager.get_panel_info(panel_id)
            if position != "central":
                dock_widget = self.create_panel(document_controller, panel_id, title, positions, position, properties)
                if dock_widget:  # could have failed to create due to exception
                    if visible_panels is None or panel_id in visible_panels:
                        dock_widget.show()
                    else:
                        dock_widget.hide()

    def create_panel(self, document_controller, panel_id, title, positions, position, properties):
        try:
            panel = self.workspace_manager.create_panel_content(panel_id, document_controller, properties)
            assert panel is not None, "panel is None [%s]" % panel_id
            assert panel.widget is not None, "panel widget is None [%s]" % panel_id
            dock_widget = document_controller.create_dock_widget(panel.widget, panel_id, title, positions, position)
            dock_widget.panel = panel
            dock_widget.on_size_changed = panel.size_changed
            dock_widget.on_focus_changed = panel.focus_changed
            dock_widget.does_retain_focus = False
            def register_ui_activity():
                self.document_controller._register_ui_activity()
            dock_widget.on_ui_activity = register_ui_activity
            self.dock_widgets.append(dock_widget)
            return dock_widget
        except Exception as e:
            import traceback
            print("Exception creating panel '" + panel_id + "': " + str(e))
            traceback.print_exc()
            traceback.print_stack()
            return None

    def display_data_item_in_display_panel(self, data_item, display_panel_id):
        for display_panel in self.display_panels:
            if display_panel.display_panel_id == display_panel_id:
                display_panel.set_display_panel_data_item(data_item)
                self.__sync_layout()

    def _construct(self, desc, display_panels, lookup_data_item):
        selected_display_panel = None
        type = desc["type"]
        container = None
        item = None
        post_children_adjust = lambda: None
        if type == "container":
            container = CanvasItem.CanvasItemComposition()
        elif type == "splitter":
            container = CanvasItem.SplitterCanvasItem(orientation=desc.get("orientation"))
            container.on_splits_will_change = functools.partial(self._splits_will_change, container)
            container.on_splits_changed = functools.partial(self._splits_did_change, container)
            def splitter_post_children_adjust():
                splits = desc.get("splits")
                if splits is not None:
                    container.splits = splits
            post_children_adjust = splitter_post_children_adjust
        elif type == "image":
            display_panel = DisplayPanel.DisplayPanel(self.document_controller, desc)
            display_panel.on_contents_changed = self.__sync_layout
            display_panels.append(display_panel)
            if desc.get("selected", False):
                selected_display_panel = display_panel
            item = display_panel
        if container:
            children = desc.get("children", list())
            for child_desc in children:
                child_canvas_item, child_selected_display_panel = self._construct(child_desc, display_panels, lookup_data_item)
                container.add_canvas_item(child_canvas_item)
                selected_display_panel = child_selected_display_panel if child_selected_display_panel else selected_display_panel
            post_children_adjust()
            return container, selected_display_panel
        return item, selected_display_panel

    def deconstruct(self) -> dict:
        return self._deconstruct(self.__canvas_item.canvas_items[0])

    def reconstruct(self, d) -> DisplayPanel.DisplayPanel:
        return self._reconstruct(d, self.__canvas_item.canvas_items[0])

    def _reconstruct(self, d, canvas_item):
        selected_display_panel = None
        type = d["type"]
        container = None
        if type == "container":
            assert isinstance(canvas_item, CanvasItem.CanvasItemComposition)
            container = canvas_item
        elif type == "splitter":
            assert isinstance(canvas_item, CanvasItem.SplitterCanvasItem)
            container = canvas_item
            splits = d.get("splits")
            if splits is not None:
                container.splits = splits
        elif type == "image":
            assert isinstance(canvas_item, DisplayPanel.DisplayPanel)
            display_panel = canvas_item
            display_panel.change_display_panel_content(d)
            if d.get("selected", False):
                selected_display_panel = display_panel
        if container:
            children = d.get("children", list())
            for child_desc, child_canvas_item in zip(children, container.canvas_items):
                child_selected_display_panel = self._reconstruct(child_desc, child_canvas_item)
                selected_display_panel = child_selected_display_panel if child_selected_display_panel else selected_display_panel
            return selected_display_panel
        return selected_display_panel

    def __get_display_panel_by_canvas_item(self, canvas_item):
        for display_panel in self.display_panels:
            if display_panel == canvas_item:
                return display_panel
        return None

    def get_display_panel_by_uuid(self, display_panel_uuid: uuid.UUID) -> typing.Optional[DisplayPanel.DisplayPanel]:
        for display_panel in self.display_panels:
            if display_panel.uuid == display_panel_uuid:
                return display_panel
        return None

    def _deconstruct(self, canvas_item):
        if isinstance(canvas_item, CanvasItem.SplitterCanvasItem):
            children = [self._deconstruct(child_canvas_item) for child_canvas_item in canvas_item.canvas_items]
            d = create_splitter_desc(canvas_item.orientation, canvas_item.splits, children)
            return d
        if isinstance(canvas_item, DisplayPanel.DisplayPanel):
            display_panel = canvas_item
            d = create_image_desc()
            if display_panel._is_selected():
                d["selected"] = True
            d.update(display_panel.save_contents())
            return d
        return None

    @property
    def _workspace(self) -> WorkspaceLayout.WorkspaceLayout:
        return self.__workspace

    @property
    def _workspace_layout(self):
        return self._deconstruct(self.__canvas_item.canvas_items[0])

    def change_workspace(self, workspace: WorkspaceLayout.WorkspaceLayout) -> None:
        command = Workspace.ChangeWorkspaceCommand(self, workspace)
        command.perform()
        self.document_controller.push_undo_command(command)

    def _change_workspace(self, workspace: WorkspaceLayout.WorkspaceLayout) -> None:
        assert workspace is not None
        # save the current workspace
        if self.__workspace:
            self.__sync_layout()
            self.__workspace = None
        # remove existing layout and canvas item
        self.display_panels = []
        for child in copy.copy(self.image_row.children):
            self.image_row.remove(child)
        # create new layout and canvas item
        self.__canvas_item = CanvasItem.CanvasItemComposition()
        canvas_widget = self.ui.create_canvas_widget()
        canvas_widget.canvas_item.add_canvas_item(self.__canvas_item)
        self._canvas_widget = canvas_widget  # only for testing
        # root canvas item should NOT be focusable or else it will grab focus in cases where no children have it.
        # no point in the root canvas item having focus.
        # self.__canvas_item.focusable = True
        # now construct the workspace
        document_model = self.document_model
        selected_display_panel = None  # avoids warning
        try:
            display_panels = list()  # to be populated by _construct
            canvas_item, selected_display_panel = self._construct(workspace.layout, display_panels, document_model.get_data_item_by_uuid)
            # store the new workspace
            if canvas_item:
                self.__workspace = workspace
                self.display_panels.extend(display_panels)
                self.__canvas_item.add_canvas_item(canvas_item)
                self.image_row.add(canvas_widget)
        except Exception as e:
            import traceback
            traceback.print_exc()
            traceback.print_stack()
        if self.__workspace == None:  # handle error condition by creating known simple workspace and replacing bad one
            d = create_image_desc()
            d["selected"] = True
            workspace.layout = d
            display_panels = list()  # to be populated by _construct
            canvas_item, selected_display_panel = self._construct(workspace.layout, display_panels, document_model.get_data_item_by_uuid)
            # store the new workspace
            if canvas_item:
                self.__workspace = workspace
                self.display_panels.extend(display_panels)
                self.__canvas_item.add_canvas_item(canvas_item)
                self.image_row.add(canvas_widget)
        self.document_controller.selected_display_panel = selected_display_panel
        document_model.workspace_uuid = workspace.uuid

    def restore(self, workspace_uuid):
        """
            Restore the workspace to the given workspace_uuid.

            If workspace_uuid is None then create a new workspace and use it.
        """
        workspace = next((workspace for workspace in self.document_model.workspaces if workspace.uuid == workspace_uuid), None)
        if workspace is None:
            workspace = self.new_workspace()
        self._change_workspace(workspace)

    def change_to_previous_workspace(self):
        workspace_uuid = self.document_model.workspace_uuid
        workspace = next((workspace for workspace in self.document_model.workspaces if workspace.uuid == workspace_uuid), None)
        workspace_index = self.document_model.workspaces.index(workspace)
        workspace_index = (workspace_index - 1) % len(self.document_model.workspaces)
        self.change_workspace(self.document_model.workspaces[workspace_index])

    def change_to_next_workspace(self):
        workspace_uuid = self.document_model.workspace_uuid
        workspace = next((workspace for workspace in self.document_model.workspaces if workspace.uuid == workspace_uuid), None)
        workspace_index = self.document_model.workspaces.index(workspace)
        workspace_index = (workspace_index + 1) % len(self.document_model.workspaces)
        self.change_workspace(self.document_model.workspaces[workspace_index])

    def new_workspace(self, name=None, layout=None, workspace_id=None, index=None) -> WorkspaceLayout.WorkspaceLayout:
        """ Create a new workspace, insert into document_model, and return it. """
        workspace = WorkspaceLayout.WorkspaceLayout()
        self.document_model.insert_workspace(index if index is not None else len(self.document_model.workspaces), workspace)
        d = create_image_desc()
        d["selected"] = True
        workspace.layout = layout if layout is not None else d
        workspace.name = name if name is not None else _("Workspace")
        if workspace_id:
            workspace.workspace_id = workspace_id
        return workspace

    def ensure_workspace(self, name, layout, workspace_id):
        """Looks for a workspace with workspace_id.

        If none is found, create a new one, add it, and change to it.
        """
        workspace = next((workspace for workspace in self.document_model.workspaces if workspace.workspace_id == workspace_id), None)
        if not workspace:
            workspace = self.new_workspace(name=name, layout=layout, workspace_id=workspace_id)
        self._change_workspace(workspace)

    def get_workspace_layout_by_uuid(self, workspace_layout_uuid: uuid.UUID) -> typing.Optional[WorkspaceLayout.WorkspaceLayout]:
        for workspace_layout in self.document_model.workspaces:
            if workspace_layout.uuid == workspace_layout_uuid:
                return workspace_layout
        return None

    class CreateWorkspaceCommand(Undo.UndoableCommand):
        def __init__(self, workspace_controller: "Workspace", name: str):
            super().__init__("Create Workspace")
            self.__workspace_controller = workspace_controller
            self.__workspace_layout_uuid = workspace_controller._workspace.uuid
            self.__new_name = name
            self.__new_layout = None
            self.__new_workspace_id = None
            self.initialize()

        def _get_modified_state(self):
            return self.__workspace_controller.document_model.modified_state

        def _set_modified_state(self, modified_state) -> None:
            self.__workspace_controller.document_model.modified_state = modified_state

        def perform(self) -> None:
            new_workspace = self.__workspace_controller.new_workspace(name=self.__new_name, layout=self.__new_layout, workspace_id=self.__new_workspace_id)
            self.__workspace_controller._change_workspace(new_workspace)

        def _undo(self) -> None:
            workspace_layout = self.__workspace_controller.get_workspace_layout_by_uuid(self.__workspace_layout_uuid)
            self.__new_layout = self.__workspace_controller._workspace.layout
            self.__new_workspace_id = self.__workspace_controller._workspace.workspace_id
            self.__workspace_controller._change_workspace(workspace_layout)

        def _redo(self) -> None:
            self.perform()

    class RemoveWorkspaceCommand(Undo.UndoableCommand):
        def __init__(self, workspace_controller: "Workspace"):
            super().__init__("Remove Workspace")
            self.__workspace_controller = workspace_controller
            self.__old_name = workspace_controller._workspace.name
            self.__old_layout = workspace_controller._workspace.layout
            self.__old_workspace_id = workspace_controller._workspace.workspace_id
            self.__old_workspace_index = workspace_controller.document_model.workspaces.index(workspace_controller._workspace)
            self.initialize()

        def _get_modified_state(self):
            return self.__workspace_controller.document_model.modified_state

        def _set_modified_state(self, modified_state) -> None:
            self.__workspace_controller.document_model.modified_state = modified_state

        def perform(self) -> None:
            old_workspace = self.__workspace_controller._workspace
            self.__workspace_controller.change_to_previous_workspace()
            self.__workspace_controller.document_model.remove_workspace(old_workspace)

        def _undo(self) -> None:
            new_workspace = self.__workspace_controller.new_workspace(name=self.__old_name, layout=self.__old_layout, workspace_id=self.__old_workspace_id, index=self.__old_workspace_index)
            self.__workspace_controller._change_workspace(new_workspace)

        def _redo(self) -> None:
            self.perform()

    class RenameWorkspaceCommand(Undo.UndoableCommand):
        def __init__(self, workspace_controller: "Workspace", name: str):
            super().__init__("Rename Workspace")
            self.__workspace_controller = workspace_controller
            self.__old_name = workspace_controller._workspace.name
            self.__new_name = name
            self.initialize()

        def _get_modified_state(self):
            return self.__workspace_controller.document_model.modified_state

        def _set_modified_state(self, modified_state) -> None:
            self.__workspace_controller.document_model.modified_state = modified_state

        def perform(self) -> None:
            self.__workspace_controller._workspace.name = self.__new_name

        def _undo(self) -> None:
            self.__workspace_controller._workspace.name = self.__old_name

        def _redo(self) -> None:
            self.perform()

    class ChangeWorkspaceCommand(Undo.UndoableCommand):
        def __init__(self, workspace_controller: "Workspace", workspace: WorkspaceLayout.WorkspaceLayout):
            super().__init__("Change Workspace")
            self.__workspace_controller = workspace_controller
            self.__old_workspace_uuid = workspace_controller._workspace.uuid
            self.__new_workspace = workspace
            self.initialize()

        def _get_modified_state(self):
            return self.__workspace_controller.document_model.modified_state

        def _set_modified_state(self, modified_state) -> None:
            self.__workspace_controller.document_model.modified_state = modified_state

        def perform(self) -> None:
            self.__workspace_controller._change_workspace(self.__new_workspace)

        def _undo(self) -> None:
            old_workspace = self.__workspace_controller.get_workspace_layout_by_uuid(self.__old_workspace_uuid)
            self.__workspace_controller._change_workspace(old_workspace)

        def _redo(self) -> None:
            self.perform()

    class ChangeContentsWorkspaceCommand(Undo.UndoableCommand):
        def __init__(self, workspace_controller: "Workspace"):
            super().__init__("Change Workspace Contents")
            self.__workspace_controller = workspace_controller
            self.__old_workspace_layout = workspace_controller.deconstruct()
            self.__new_workspace_layout = None
            self.initialize()

        def _get_modified_state(self):
            return self.__workspace_controller.document_model.modified_state

        def _set_modified_state(self, modified_state) -> None:
            self.__workspace_controller.document_model.modified_state = modified_state

        def _undo(self) -> None:
            self.__new_workspace_layout = self.__workspace_controller.deconstruct()
            self.__workspace_controller.reconstruct(self.__old_workspace_layout)

        def _redo(self) -> None:
            self.__workspace_controller.reconstruct(self.__new_workspace_layout)

    def create_workspace(self) -> None:
        """ Pose a dialog to name and create a workspace. """

        def create_clicked(text):
            if text:
                command = Workspace.CreateWorkspaceCommand(self, text)
                command.perform()
                self.document_controller.push_undo_command(command)

        self.pose_get_string_message_box(caption=_("Enter a name for the workspace"), text=_("Workspace"),
                                         accepted_fn=create_clicked, accepted_text=_("Create"),
                                         message_box_id="create_workspace")

    def rename_workspace(self) -> None:
        """ Pose a dialog to rename the workspace. """

        def rename_clicked(text):
            if len(text) > 0:
                command = Workspace.RenameWorkspaceCommand(self, text)
                command.perform()
                self.document_controller.push_undo_command(command)

        self.pose_get_string_message_box(caption=_("Enter new name for workspace"), text=self.__workspace.name,
                                         accepted_fn=rename_clicked, accepted_text=_("Rename"),
                                         message_box_id="rename_workspace")

    def remove_workspace(self):
        """ Pose a dialog to confirm removal then remove workspace. """

        def confirm_clicked():
            if len(self.document_model.workspaces) > 1:
                command = Workspace.RemoveWorkspaceCommand(self)
                command.perform()
                self.document_controller.push_undo_command(command)

        caption = _("Remove workspace named '{0}'?").format(self.__workspace.name)
        self.pose_confirmation_message_box(caption, confirm_clicked, accepted_text=_("Remove Workspace"),
                                           message_box_id="remove_workspace")

    def pose_get_string_message_box(self, caption, text, accepted_fn, rejected_fn=None, accepted_text=None, rejected_text=None, message_box_id=None):
        message_box_id = message_box_id if message_box_id else str(uuid.uuid4())
        if message_box_id in self.__message_boxes:
            return None
        if accepted_text is None: accepted_text = _("OK")
        if rejected_text is None: rejected_text = _("Cancel")
        message_box_widget = self.ui.create_column_widget()  # properties={"stylesheet": "background: #FFD"}
        caption_row = self.ui.create_row_widget()
        caption_row.add_spacing(12)
        caption_row.add(self.ui.create_label_widget(caption))
        caption_row.add_stretch()
        inside_row = self.ui.create_row_widget()

        def reject_button_clicked():
            if rejected_fn: rejected_fn()
            async def remove_widget():
                self.message_column.remove(message_box_widget)
            self.document_controller.event_loop.create_task(remove_widget())
            del self.__message_boxes[message_box_id]
            return False

        def accept_button_clicked():
            accepted_fn(string_edit_widget.text)
            async def remove_widget():
                self.message_column.remove(message_box_widget)
            self.document_controller.event_loop.create_task(remove_widget())
            del self.__message_boxes[message_box_id]
            return False

        string_edit_widget = self.ui.create_line_edit_widget()
        string_edit_widget.text = text
        string_edit_widget.on_return_pressed = accept_button_clicked
        string_edit_widget.on_escape_pressed = reject_button_clicked
        reject_button = self.ui.create_push_button_widget(rejected_text)
        reject_button.on_clicked = reject_button_clicked
        accepted_button = self.ui.create_push_button_widget(accepted_text)
        accepted_button.on_clicked = accept_button_clicked
        inside_row.add_spacing(12)
        inside_row.add(string_edit_widget)
        inside_row.add_spacing(12)
        inside_row.add(reject_button)
        inside_row.add_spacing(12)
        inside_row.add(accepted_button)
        inside_row.add_stretch()
        message_box_widget.add_spacing(6)
        message_box_widget.add(caption_row)
        message_box_widget.add_spacing(4)
        message_box_widget.add(inside_row)
        message_box_widget.add_spacing(4)
        self.message_column.add(message_box_widget)
        string_edit_widget.select_all()
        string_edit_widget.focused = True
        self.__message_boxes[message_box_id] = message_box_widget
        return message_box_widget

    def pose_confirmation_message_box(self, caption, accepted_fn, rejected_fn=None, accepted_text=None, rejected_text=None, display_rejected=True, message_box_id=None):
        message_box_id = message_box_id if message_box_id else str(uuid.uuid4())
        if message_box_id in self.__message_boxes:
            return None
        if accepted_text is None: accepted_text = _("OK")
        if rejected_text is None: rejected_text = _("Cancel")
        message_box_widget = self.ui.create_column_widget()  # properties={"stylesheet": "background: #FFD"}

        def reject_button_clicked():
            if rejected_fn: rejected_fn()
            self.message_column.remove(message_box_widget)
            del self.__message_boxes[message_box_id]

        def accept_button_clicked():
            accepted_fn()
            self.message_column.remove(message_box_widget)
            del self.__message_boxes[message_box_id]

        reject_button = self.ui.create_push_button_widget(rejected_text)
        reject_button.on_clicked = reject_button_clicked
        accepted_button = self.ui.create_push_button_widget(accepted_text)
        accepted_button.on_clicked = accept_button_clicked
        caption_row = self.ui.create_row_widget()
        caption_row.add_spacing(12)
        caption_row.add(self.ui.create_label_widget(caption))
        if display_rejected:
            caption_row.add_spacing(12)
            caption_row.add(reject_button)
        caption_row.add_spacing(12)
        caption_row.add(accepted_button)
        caption_row.add_stretch()
        message_box_widget.add_spacing(6)
        message_box_widget.add(caption_row)
        message_box_widget.add_spacing(4)
        self.message_column.add(message_box_widget)
        self.__message_boxes[message_box_id] = message_box_widget
        return message_box_widget

    def handle_drag_enter(self, display_panel, mime_data):
        if mime_data.has_format("text/library_item_uuid"):
            return "copy"
        if mime_data.has_format("text/data_item_uuid"):
            return "copy"
        if mime_data.has_format("text/uri-list"):
            return "copy"
        if mime_data.has_format(DisplayPanel.DISPLAY_PANEL_MIME_TYPE):
            return "copy"
        return "ignore"

    def handle_drag_leave(self, display_panel):
        return False

    def handle_drag_move(self, display_panel, mime_data, x, y):
        if mime_data.has_format("text/library_item_uuid"):
            return "copy"
        if mime_data.has_format("text/data_item_uuid"):
            return "copy"
        if mime_data.has_format("text/uri-list"):
            return "copy"
        if mime_data.has_format(DisplayPanel.DISPLAY_PANEL_MIME_TYPE):
            return "copy"
        return "ignore"

    def handle_drop(self, display_panel, mime_data, region, x, y):
        document_model = self.document_model
        if mime_data.has_format(DisplayPanel.DISPLAY_PANEL_MIME_TYPE):
            d = json.loads(mime_data.data_as_string(DisplayPanel.DISPLAY_PANEL_MIME_TYPE))
            if region == "right" or region == "left" or region == "top" or region == "bottom":
                command = self.insert_display_panel(display_panel, region, None, d)
                self.document_controller.push_undo_command(command)
            else:
                command = self.__replace_displayed_data_item(display_panel, None, d)
                self.document_controller.push_undo_command(command)
            return "move"
        if mime_data.has_format("text/library_item_uuid"):
            library_item_uuid = uuid.UUID(mime_data.data_as_string("text/library_item_uuid"))
            library_item = document_model.get_data_item_by_key(library_item_uuid)
            if library_item:
                if region == "right" or region == "left" or region == "top" or region == "bottom":
                    command = self.insert_display_panel(display_panel, region, library_item)
                    self.document_controller.push_undo_command(command)
                else:
                    command = self.__replace_displayed_data_item(display_panel, library_item)
                    self.document_controller.push_undo_command(command)
                return "copy"
        if mime_data.has_format("text/data_item_uuid"):
            data_item_uuid = uuid.UUID(mime_data.data_as_string("text/data_item_uuid"))
            data_item = document_model.get_data_item_by_key(data_item_uuid)
            if data_item:
                if region == "right" or region == "left" or region == "top" or region == "bottom":
                    command = self.insert_display_panel(display_panel, region, data_item)
                    self.document_controller.push_undo_command(command)
                else:
                    command = self.__replace_displayed_data_item(display_panel, data_item)
                    self.document_controller.push_undo_command(command)
                return "copy"
        if mime_data.has_format("text/uri-list"):
            index = len(document_model.data_items)
            self.document_controller.receive_files(mime_data.file_paths, None, index, threaded=True, display_panel=display_panel)
            return "copy"
        return "ignore"

    def _replace_displayed_data_item(self, display_panel, data_item, d=None) -> Undo.UndoableCommand:
        return self.__replace_displayed_data_item(display_panel, data_item, d)

    def __replace_displayed_data_item(self, display_panel, data_item, d=None) -> Undo.UndoableCommand:
        """ Used in drag/drop support. """
        self.document_controller.replaced_display_panel_content = display_panel.save_contents()
        command = DisplayPanel.ReplaceDisplayPanelCommand(self)
        if data_item:
            display_panel.set_display_panel_data_item(data_item, detect_controller=True)
        elif d is not None:
            display_panel.change_display_panel_content(d)
        display_panel.request_focus()
        self.__sync_layout()
        return command

    class SplitDisplayPanelCommand(Undo.UndoableCommand):
        def __init__(self, workspace_controller, workspace_layout, modified_state, display_panel, region_id: str, data_item, d: dict, old_splits, new_display_panel):
            super().__init__("Split Display Panel")
            self.__workspace_controller = workspace_controller
            self.__workspace_layout_uuid = workspace_layout.uuid
            self.__display_panel_uuid = display_panel.uuid
            self.__region_id = region_id
            self.__data_item_uuid = data_item.uuid if data_item else None
            self.__d = d
            self.__old_splits = old_splits
            self.__new_display_panel_uuid = new_display_panel.uuid
            self.__uuid = new_display_panel.uuid
            self.initialize(modified_state)

        def close(self):
            self.__workspace_controller = None
            self.__workspace_layout_uuid = None
            self.__display_panel_uuid = None
            self.__region_id = None
            self.__data_item_uuid = None
            self.__d = None
            self.__old_splits = None
            self.__new_display_panel_uuid = None
            self.__uuid = None
            super().close()

        def _get_modified_state(self):
            workspace_layout = self.__workspace_controller.get_workspace_layout_by_uuid(self.__workspace_layout_uuid)
            return workspace_layout.modified_state

        def _set_modified_state(self, modified_state) -> None:
            workspace_layout = self.__workspace_controller.get_workspace_layout_by_uuid(self.__workspace_layout_uuid)
            workspace_layout.modified_state = modified_state

        @property
        def _old_splits(self):
            return self.__old_splits

        def _undo(self) -> None:
            new_display_panel = self.__workspace_controller.get_display_panel_by_uuid(self.__new_display_panel_uuid)
            self.__d = new_display_panel.save_contents()
            self.__workspace_controller._remove_display_panel(new_display_panel, self.__old_splits)
            self.__data_item_uuid = None

        def _redo(self) -> None:
            data_item = self.__workspace_controller.document_model.get_data_item_by_uuid(self.__data_item_uuid)
            display_panel = self.__workspace_controller.get_display_panel_by_uuid(self.__display_panel_uuid)
            _, new_display_panel = self.__workspace_controller._insert_display_panel(display_panel, self.__region_id, data_item, self.__d, self.__uuid)
            self.__new_display_panel_uuid = new_display_panel.uuid

    class RemoveDisplayPanelCommand(Undo.UndoableCommand):

        def __init__(self, workspace_controller, workspace_layout, modified_state, old_display_panel, region, d, old_uuid, old_splits):
            super().__init__("Remove Display Panel")
            self.__workspace_controller = workspace_controller
            self.__workspace_layout_uuid = workspace_layout.uuid
            self.__old_display_panel = old_display_panel  # canvas_item: always valid even in redo/undo
            self.__region = region
            self.__d = d
            self.__old_uuid = old_uuid
            self.__old_splits = old_splits
            self.__new_splits = None
            self.__new_display_panel_uuid = None
            self.initialize(modified_state)

        def close(self):
            self.__workspace_controller = None
            self.__workspace_layout_uuid = None
            self.__old_display_panel = None
            self.__region = None
            self.__d = None
            self.__old_uuid = None
            self.__old_splits = None
            self.__new_splits = None
            self.__new_display_panel_uuid = None
            super().close()

        def _get_modified_state(self):
            workspace_layout = self.__workspace_controller.get_workspace_layout_by_uuid(self.__workspace_layout_uuid)
            return workspace_layout.modified_state

        def _set_modified_state(self, modified_state) -> None:
            workspace_layout = self.__workspace_controller.get_workspace_layout_by_uuid(self.__workspace_layout_uuid)
            workspace_layout.modified_state = modified_state

        def _undo(self) -> None:
            old_splits, new_display_panel = self.__workspace_controller._insert_display_panel(self.__old_display_panel, self.__region, None, self.__d, self.__old_uuid, self.__old_splits)
            self.__new_splits = old_splits
            self.__new_display_panel_uuid = new_display_panel.uuid

        def _redo(self) -> None:
            new_display_panel = self.__workspace_controller.get_display_panel_by_uuid(self.__new_display_panel_uuid)
            old_display_panel, _, _ = self.__workspace_controller._remove_display_panel(new_display_panel, self.__new_splits)
            self.__old_display_panel = old_display_panel

    def insert_display_panel(self, display_panel, region, data_item=None, d=None, new_uuid=None, new_splits=None) -> Undo.UndoableCommand:
        modified_state = self.__workspace.modified_state
        old_splits, new_display_panel = self._insert_display_panel(display_panel, region, data_item, d, new_uuid, new_splits)
        return Workspace.SplitDisplayPanelCommand(self, self.__workspace, modified_state, display_panel, region, data_item, d, old_splits, new_display_panel)

    def _insert_display_panel(self, display_panel, region, data_item, d, new_uuid, new_splits=None) -> typing.Tuple[typing.Optional[list], typing.Optional[DisplayPanel.DisplayPanel]]:
        assert isinstance(display_panel, DisplayPanel.DisplayPanel)
        orientation = "vertical" if region in ("left", "right") else "horizontal"
        new_display_panel = None
        container = display_panel.container
        # record old splits for undo if modifying an existing splitter; otherwise removing the display panel is trivial
        old_splits = list(container.splits) if isinstance(container, CanvasItem.SplitterCanvasItem) and container.orientation == orientation else None
        # if not modifying an existing splitter or if modifying in the other orientation, wrap panel in new splitter
        if not isinstance(container, CanvasItem.SplitterCanvasItem) or container.orientation != orientation:
            # check if trying to drag on non-axis edge of splitter
            # special case where top level item is the image panel
            splitter_canvas_item = CanvasItem.SplitterCanvasItem(orientation=orientation)
            splitter_canvas_item.on_splits_will_change = functools.partial(self._splits_will_change, splitter_canvas_item)
            splitter_canvas_item.on_splits_changed = functools.partial(self._splits_did_change, splitter_canvas_item)
            container.wrap_canvas_item(display_panel, splitter_canvas_item)
            container = splitter_canvas_item
        index = container.canvas_items.index(display_panel)
        if isinstance(container, CanvasItem.SplitterCanvasItem):
            # modify the existing splitter
            old_split = container.splits[index] if not new_splits else None
            new_index_adj = 1 if region == "right" or region == "bottom" else 0
            new_display_panel = DisplayPanel.DisplayPanel(self.document_controller, dict(), new_uuid)
            self.display_panels.insert(self.display_panels.index(display_panel) + new_index_adj, new_display_panel)
            if data_item:
                new_display_panel.set_display_panel_data_item(data_item, detect_controller=True)
            elif d is not None:
                new_display_panel.change_display_panel_content(d)
            container.insert_canvas_item(index + new_index_adj, new_display_panel)
            self.document_controller.selected_display_panel = new_display_panel
            # adjust the splits
            if new_splits:
                container.splits = copy.copy(new_splits)
            else:
                splits = list(container.splits)
                splits[index] = old_split * 0.5
                splits[index + 1] = old_split * 0.5
                container.splits = splits
            new_display_panel.request_focus()
        self.__sync_layout()
        return old_splits, new_display_panel

    def remove_display_panel(self, display_panel, splits=None) -> Undo.UndoableCommand:
        # save the old display panel
        d = display_panel.save_contents()
        modified_state = self.__workspace.modified_state
        old_display_panel, old_splits, region_id = self._remove_display_panel(display_panel, splits)
        return Workspace.RemoveDisplayPanelCommand(self, self.__workspace, modified_state, old_display_panel, region_id, d, display_panel.uuid, old_splits)

    def _remove_display_panel(self, display_panel, splits):
        # first make sure the display panel has no content
        display_panel.change_display_panel_content({"type": "image", "display-panel-type": "empty-display-panel"})
        # now remove it
        container = display_panel.container
        region_id = None
        old_display_panel = None
        old_splits = None
        if isinstance(container, CanvasItem.SplitterCanvasItem):
            old_splits = container.splits
            if display_panel in container.canvas_items:
                if len(container.canvas_items) > 1:
                    # configure the redo
                    display_panel_index = container.canvas_items.index(display_panel)
                    if display_panel_index > 0:
                        old_display_panel = container.canvas_items[display_panel_index - 1]
                        region_id = "right" if container.orientation == "vertical" else "bottom"
                    else:
                        old_display_panel = container.canvas_items[1]
                        region_id = "left" if container.orientation == "vertical" else "top"
                self.display_panels.remove(display_panel)
                container.remove_canvas_item(display_panel)
                if len(container.canvas_items) == 1:
                    container.unwrap_canvas_item(container.canvas_items[0])
                else:
                    if splits:
                        container.splits = copy.copy(splits)
        self.__sync_layout()
        return old_display_panel, old_splits, region_id

    def selected_display_panel_changed(self, selected_display_panel):
        for display_panel in self.display_panels:
            display_panel.set_selected(display_panel == selected_display_panel)

    def _splits_will_change(self, splitter_canvas_item):
        self.__change_splitter_splits = splitter_canvas_item.splits
        self.__change_splitter_command = Workspace.ChangeContentsWorkspaceCommand(self)

    def _splits_did_change(self, splitter_canvas_item):
        self.__sync_layout()
        if self.__change_splitter_command:
            if splitter_canvas_item.splits != self.__change_splitter_splits:
                self.document_controller.push_undo_command(self.__change_splitter_command)
            else:
                self.__change_splitter_command.close()
            self.__change_splitter_command = None
            self.__change_splitter_splits = None

    def __sync_layout(self):
        # ensure that the layout is written to persistent storage
        self.__workspace.layout = self._deconstruct(self.__canvas_item.canvas_items[0])


class WorkspaceManager(metaclass=Utility.Singleton):

    """
        The WorkspaceManager object keeps a list of workspaces and a list of panel
        types. It also creates workspace objects.
        """
    def __init__(self):
        self.__panel_tuples = {}

    def register_panel(self, panel_class, panel_id, name, positions, position, properties=None):
        panel_tuple = panel_class, panel_id, name, positions, position, properties
        self.__panel_tuples[panel_id] = panel_tuple

    def unregister_panel(self, panel_id):
        del self.__panel_tuples[panel_id]

    def create_panel_content(self, panel_id, document_controller, properties=None):
        if panel_id in self.__panel_tuples:
            tuple = self.__panel_tuples[panel_id]
            cls = tuple[0]
            try:
                properties = properties if properties else {}
                panel = cls(document_controller, panel_id, properties)
                return panel
            except Exception as e:
                import traceback
                print("Exception creating panel '" + panel_id + "': " + str(e))
                traceback.print_exc()
                traceback.print_stack()
        return None

    def register_filter_panel(self, filter_panel_class):
        self.__filter_panel_class = filter_panel_class

    def create_filter_panel(self, document_controller):
        return self.__filter_panel_class(document_controller) if self.__filter_panel_class else None

    def get_panel_info(self, panel_id):
        assert panel_id in self.__panel_tuples
        tuple = self.__panel_tuples[panel_id]
        return tuple[2], tuple[3], tuple[4], tuple[5]

    def __get_panel_ids(self):
        return list(self.__panel_tuples.keys())
    panel_ids = property(__get_panel_ids)



# standard libraries
import abc
import copy
import datetime
import functools
import gettext
import os
import threading
import time
import typing
import uuid
import warnings
import weakref

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.data import Core
from nion.data import DataAndMetadata
from nion.data import Image
from nion.swift.model import Cache
from nion.swift.model import Connection
from nion.swift.model import Display
from nion.swift.model import Graphics
from nion.swift.model import Metadata
from nion.swift.model import Symbolic
from nion.swift.model import Utility
from nion.utils import Converter
from nion.utils import Event
from nion.utils import Observable
from nion.utils import Persistence

_ = gettext.gettext

UNTITLED_STR = _("Untitled")


class CalibrationList:

    def __init__(self, calibrations=None):
        self.list = list() if calibrations is None else copy.deepcopy(calibrations)

    def read_dict(self, storage_list):
        # storage_list will be whatever is returned by write_dict.
        new_list = list()
        for calibration_dict in storage_list:
            new_list.append(Calibration.Calibration().read_dict(calibration_dict))
        self.list = new_list
        return self  # for convenience

    def write_dict(self):
        list = []
        for calibration in self.list:
            list.append(calibration.write_dict())
        return list


"""
    Data sources are interfaces to get data and metadata.

    Data sources should support deep copy. They are never shared (i.e. they only occur once in the model).

    When data sources is about to be closed, they will receive this message.

    closed()

    When data sources are about to be removed, they will receive this message.

    about_to_be_removed()

    Data sources can only be associated in the operation tree of a single data item. The data source
    keeps track of that so as to maintain the dependent_data_item list on data items.

    set_dependent_data_item(data_item)

    Data sources can subscribe to other data items becoming available via the data item manager.
"""


# enumerations for types of data item content changes
DATA = 1
METADATA = 2
DISPLAYS = 3


class DtypeToStringConverter:
    def convert(self, value):
        return str(value) if value is not None else None
    def convert_back(self, value):
        return numpy.dtype(value) if value is not None else None


class DatetimeToStringConverter:
    def convert(self, value):
        return value.isoformat() if value is not None else None
    def convert_back(self, value):
        try:
            if len(value) == 26:
                return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f")
            elif len(value) == 19:
                return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")
        except ValueError as e:
            pass  # fall through
        return None


def data_source_factory(lookup_id):
    type = lookup_id("type")
    if type == "buffered-data-source":
        return BufferedDataSource()
    else:
        return None


class UuidsToStringsConverter:
    def convert(self, value):
        return [str(uuid_) for uuid_ in value]
    def convert_back(self, value):
        return [uuid.UUID(uuid_str) for uuid_str in value]


class BufferedDataSource(Observable.Observable, Persistence.PersistentObject):

    """
    A data source that stores the data directly, with optional source data that gets updated as necessary.

    The buffered data source stores data directly. If the optional data_source is present, it will update the stored
    data when the data_source is updated.

    Coordinate system. The coordinate system of the pixels refers to the position within the numpy array. For 1d data,
    this means that channel 0 is the first channel. For 2d data, this means that the pixel coordinate 0, 0 is at the top
    left, within increasing y moving downward and increasing x moving right. For 3d data, this means that the first
    coordinate specifies the depth with 0 considered to be the "top". The next two coordinates are y, x with 0, 0 at the
    top left of each layer.
    """

    def __init__(self, data=None):
        super().__init__()
        self.define_type("buffered-data-source")
        self.persistent_storage = None
        data_shape = data.shape if data is not None else None
        data_dtype = data.dtype if data is not None else None
        # windows utcnow has a resolution of 1ms, this sleep can guarantee unique times for all created times during a particular test.
        # this is not my favorite solution since it limits data item creation to 1000/s but until I find a better solution, this is my compromise.
        self.define_property("data_shape", data_shape, hidden=True, recordable=False)
        self.define_property("data_dtype", data_dtype, hidden=True, recordable=False, converter=DtypeToStringConverter())
        self.define_property("is_sequence", False, recordable=False, changed=self.__data_description_changed)
        dimensional_shape = Image.dimensional_shape_from_shape_and_dtype(data_shape, data_dtype)
        collection_dimension_count = (2 if len(dimensional_shape) == 3 else 0) if dimensional_shape is not None else None
        datum_dimension_count = len(dimensional_shape) - collection_dimension_count if dimensional_shape is not None else None
        self.define_property("collection_dimension_count", collection_dimension_count, recordable=False, changed=self.__data_description_changed)
        self.define_property("datum_dimension_count", datum_dimension_count, recordable=False, changed=self.__data_description_changed)
        self.define_property("intensity_calibration", Calibration.Calibration(), hidden=True, make=Calibration.Calibration, changed=self.__metadata_property_changed)
        self.define_property("dimensional_calibrations", CalibrationList(), hidden=True, make=CalibrationList, changed=self.__dimensional_calibrations_changed)
        self.define_property("created", datetime.datetime.utcnow(), converter=DatetimeToStringConverter(), changed=self.__metadata_property_changed)
        self.define_property("data_modified", recordable=False, converter=DatetimeToStringConverter(), changed=self.__metadata_property_changed)
        self.define_property("metadata", dict(), hidden=True, changed=self.__metadata_property_changed)
        self.__timezone = None  # set by the data item, used when returning data_and_metadata
        self.__timezone_offset = None  # set by the data item, used when returning data_and_metadata
        self.__data_and_metadata = None
        self.__data_and_metadata_lock = threading.RLock()
        self.__intensity_calibration = None
        self.__dimensional_calibrations = list()
        self.__metadata = dict()
        self.__data_ref_count = 0
        self.__data_ref_count_mutex = threading.RLock()
        self.data_changed_event = Event.Event()
        self.metadata_changed_event = Event.Event()
        if data is not None:
            data_and_metadata = DataAndMetadata.DataAndMetadata.from_data(data)
            self.__set_data_metadata_direct(data_and_metadata)
        self._about_to_be_removed = False
        self._closed = False

    def close(self):
        self.__data_and_metadata = None
        assert self._about_to_be_removed
        assert not self._closed
        self._closed = True

    def about_to_be_removed(self):
        # called before close and before item is removed from its container
        assert not self._about_to_be_removed
        self._about_to_be_removed = True

    def __deepcopy__(self, memo):
        deepcopy = self.__class__()
        deepcopy.deepcopy_from(self, memo)
        memo[id(self)] = deepcopy
        return deepcopy

    def deepcopy_from(self, buffered_data_source, memo):
        super().deepcopy_from(buffered_data_source, memo)
        # data and metadata
        self.set_data_and_metadata(copy.deepcopy(buffered_data_source.data_and_metadata))

    def clone(self) -> "BufferedDataSource":
        data_source = BufferedDataSource()
        data_source.uuid = self.uuid
        return data_source

    def snapshot(self):
        """
            Take a snapshot and return a new buffered data source. A snapshot is a copy of everything
            except the data and operation which are replaced by new data with the operation
            applied or "burned in".
        """
        buffered_data_source_copy = BufferedDataSource()
        buffered_data_source_copy.set_data_and_metadata(copy.deepcopy(self.data_and_metadata))
        return buffered_data_source_copy

    def read_from_dict(self, properties):
        super().read_from_dict(properties)
        data_shape = self._get_persistent_property_value("data_shape", None)
        data_dtype = DtypeToStringConverter().convert_back(self._get_persistent_property_value("data_dtype", None))
        if data_shape is not None and data_dtype is not None:
            data_shape_and_dtype = data_shape, data_dtype
            dimensional_shape = Image.dimensional_shape_from_shape_and_dtype(data_shape, data_dtype)
            intensity_calibration = self._get_persistent_property_value("intensity_calibration")
            dimensional_calibration_list = self._get_persistent_property_value("dimensional_calibrations")
            dimensional_calibrations = dimensional_calibration_list.list if dimensional_calibration_list else None
            if dimensional_calibrations is not None:
                dimensional_shape = Image.dimensional_shape_from_shape_and_dtype(data_shape, data_dtype)
                while len(dimensional_shape) > len(dimensional_calibrations):
                    dimensional_calibrations.append(Calibration.Calibration())
                while len(dimensional_shape) < len(dimensional_calibrations):
                    dimensional_calibrations.pop(-1)
            metadata = self._get_persistent_property_value("metadata")
            timestamp = self._get_persistent_property_value("created")
            if timestamp is None:  # invalid timestamp -- set property to now but don't trigger change
                timestamp = datetime.datetime.now()
                self._get_persistent_property("created").value = timestamp
            is_sequence = self._get_persistent_property_value("is_sequence", False)
            collection_dimension_count = self._get_persistent_property_value("collection_dimension_count")
            datum_dimension_count = self._get_persistent_property_value("datum_dimension_count")
            if collection_dimension_count is None:
                collection_dimension_count = 2 if len(dimensional_shape) == 3 and not is_sequence else 0
                # update collection_dimension_count, but in a way that sets the internal value but
                # doesn't trigger a write to disk or a change modification.
                self._get_persistent_property("collection_dimension_count").set_value(collection_dimension_count)
            if datum_dimension_count is None:
                datum_dimension_count = len(dimensional_shape) - collection_dimension_count - (1 if is_sequence else 0)
                # update collection_dimension_count, but in a way that sets the internal value but
                # doesn't trigger a write to disk or a change modification.
                self._get_persistent_property("datum_dimension_count").set_value(datum_dimension_count)
            data_descriptor = DataAndMetadata.DataDescriptor(is_sequence, collection_dimension_count, datum_dimension_count)
            self.__data_and_metadata = DataAndMetadata.DataAndMetadata(self.__load_data, data_shape_and_dtype, intensity_calibration, dimensional_calibrations, metadata, timestamp,
                                                                       data_descriptor=data_descriptor, timezone=self.timezone, timezone_offset=self.timezone_offset)
            with self.__data_ref_count_mutex:
                self.__data_and_metadata._add_data_ref_count(self.__data_ref_count)
            self.__data_and_metadata.unloadable = self.persistent_object_context is not None

    def persistent_object_context_changed(self):
        super().persistent_object_context_changed()
        if self.__data_and_metadata:
            self.__data_and_metadata.unloadable = self.persistent_object_context is not None

    def __data_description_changed(self, name, value):
        self.__property_changed(name, value)
        self.__metadata_changed()

    def __dimensional_calibrations_changed(self, name, value):
        self.__property_changed(name, self.dimensional_calibrations)  # don't send out the CalibrationList object
        self.__metadata_changed()

    def __metadata_property_changed(self, name, value):
        self.__property_changed(name, value)
        self.__metadata_changed()

    def __metadata_changed(self):
        self.metadata_changed_event.fire()

    def __property_changed(self, name, value):
        self.notify_property_changed(name)

    @property
    def timezone(self):
        return self.__timezone

    @timezone.setter
    def timezone(self, value):
        self.__timezone = value
        if self.__data_and_metadata:
            self.__data_and_metadata.data_metadata.timezone = self.__timezone

    @property
    def timezone_offset(self):
        return self.__timezone_offset

    @timezone_offset.setter
    def timezone_offset(self, value):
        self.__timezone_offset = value
        if self.__data_and_metadata:
            self.__data_and_metadata.data_metadata.timezone_offset = self.__timezone_offset

    @property
    def data_metadata(self):
        return self.__data_and_metadata.data_metadata if self.__data_and_metadata else None

    @property
    def data_and_metadata(self):
        return self.__data_and_metadata

    def __load_data(self):
        if self.persistent_storage:
            return self.persistent_storage.load_data()
        return None

    def __set_data_metadata_direct(self, data_and_metadata, data_modified=None):
        self.__data_and_metadata = data_and_metadata
        if self.__data_and_metadata:
            with self.__data_ref_count_mutex:
                self.__data_and_metadata._add_data_ref_count(self.__data_ref_count)
        if self.__data_and_metadata:
            self._set_persistent_property_value("data_shape", self.__data_and_metadata.data_shape)
            self._set_persistent_property_value("data_dtype", DtypeToStringConverter().convert(self.__data_and_metadata.data_dtype))
            self._set_persistent_property_value("is_sequence", self.__data_and_metadata.is_sequence)
            self._set_persistent_property_value("collection_dimension_count", self.__data_and_metadata.collection_dimension_count)
            self._set_persistent_property_value("datum_dimension_count", self.__data_and_metadata.datum_dimension_count)
            self._set_persistent_property_value("intensity_calibration", self.__data_and_metadata.intensity_calibration)
            self._set_persistent_property_value("dimensional_calibrations", CalibrationList(self.__data_and_metadata.dimensional_calibrations))
            metadata_copy = copy.deepcopy(self.__data_and_metadata.metadata)
            self.__metadata = metadata_copy
            self._set_persistent_property_value("metadata", metadata_copy)
        self.data_modified = data_modified if data_modified else datetime.datetime.utcnow()
        self.data_changed_event.fire(self)

    def set_data_and_metadata(self, data_and_metadata, data_modified=None):
        """Sets the underlying data and data-metadata to the data_and_metadata.

        Note: this does not make a copy of the data.
        """
        self.increment_data_ref_count()
        try:
            if data_and_metadata:
                data = data_and_metadata.data
                data_shape_and_dtype = data_and_metadata.data_shape_and_dtype
                intensity_calibration = data_and_metadata.intensity_calibration
                dimensional_calibrations = data_and_metadata.dimensional_calibrations
                metadata = data_and_metadata.metadata
                timestamp = data_and_metadata.timestamp
                data_descriptor = data_and_metadata.data_descriptor
                new_data_and_metadata = DataAndMetadata.DataAndMetadata(self.__load_data, data_shape_and_dtype, intensity_calibration, dimensional_calibrations, metadata, timestamp, data, data_descriptor)
            else:
                new_data_and_metadata = None
            self.__set_data_metadata_direct(new_data_and_metadata, data_modified)
            if self.__data_and_metadata is not None:
                if self.persistent_storage:
                    self.persistent_storage.update_data(self.__data_and_metadata.data)
                    self.__data_and_metadata.unloadable = True
        finally:
            self.decrement_data_ref_count()

    @property
    def dimensional_shape(self):
        return self.__data_and_metadata.dimensional_shape if self.__data_and_metadata else list()

    @property
    def is_collection(self):
        return self.collection_dimension_count > 0 if self.collection_dimension_count is not None else False

    @property
    def intensity_calibration(self):
        return copy.deepcopy(self.__data_and_metadata.intensity_calibration) if self.__data_and_metadata else self.__intensity_calibration

    @intensity_calibration.setter
    def intensity_calibration(self, intensity_calibration):
        """ Set the intensity calibration. """
        if self.__data_and_metadata:  # handle case of missing data and metadata but doing recording
            self.__data_and_metadata._set_intensity_calibration(intensity_calibration)
        self.__intensity_calibration = copy.deepcopy(intensity_calibration)  # backup in case of no data and metadata
        self._set_persistent_property_value("intensity_calibration", intensity_calibration)

    def set_intensity_calibration(self, intensity_calibration):
        self.intensity_calibration = intensity_calibration

    @property
    def dimensional_calibrations(self):
        return copy.deepcopy(self.__data_and_metadata.dimensional_calibrations) if self.__data_and_metadata else self.__dimensional_calibrations

    @dimensional_calibrations.setter
    def dimensional_calibrations(self, dimensional_calibrations):
        """ Set the dimensional calibrations. """
        if self.__data_and_metadata:  # handle case of missing data and metadata but doing recording
            self.__data_and_metadata._set_dimensional_calibrations(dimensional_calibrations)
        self.__dimensional_calibrations = copy.deepcopy(dimensional_calibrations)  # backup in case of no data and metadata
        self._set_persistent_property_value("dimensional_calibrations", CalibrationList(dimensional_calibrations))

    def set_dimensional_calibrations(self, dimensional_calibrations):
        self.dimensional_calibrations = dimensional_calibrations

    def set_dimensional_calibration(self, dimension, calibration):
        dimensional_calibrations = self.dimensional_calibrations
        while len(dimensional_calibrations) <= dimension:
            dimensional_calibrations.append(Calibration.Calibration())
        dimensional_calibrations[dimension] = calibration
        self.set_dimensional_calibrations(dimensional_calibrations)

    @property
    def metadata(self):
        return copy.deepcopy(self.__data_and_metadata.metadata) if self.__data_and_metadata else self.__metadata

    @metadata.setter
    def metadata(self, metadata):
        assert isinstance(metadata, dict)
        self.__data_and_metadata._set_metadata(metadata)
        self.__metadata = copy.deepcopy(metadata)
        self._set_persistent_property_value("metadata", self.__metadata)

    def _get_data(self):
        return self.__data_and_metadata.data if self.__data_and_metadata else None

    def _update_data(self, data, data_modified=None):
        dimensional_shape = Image.dimensional_shape_from_data(data)
        data_and_metadata = self.data_and_metadata
        intensity_calibration = data_and_metadata.intensity_calibration if data_and_metadata else None
        dimensional_calibrations = copy.deepcopy(data_and_metadata.dimensional_calibrations) if data_and_metadata else None
        if data_and_metadata:
            while len(dimensional_calibrations) < len(dimensional_shape):
                dimensional_calibrations.append(Calibration.Calibration())
            while len(dimensional_calibrations) > len(dimensional_shape):
                dimensional_calibrations.pop(-1)
        metadata = data_and_metadata.metadata if data_and_metadata else None
        timestamp = None  # always update when the data is modified
        self.set_data_and_metadata(DataAndMetadata.DataAndMetadata.from_data(data, intensity_calibration, dimensional_calibrations, metadata, timestamp), data_modified)

    def increment_data_ref_count(self):
        with self.__data_ref_count_mutex:
            initial_count = self.__data_ref_count
            self.__data_ref_count += 1
            if initial_count == 0 and self.__data_and_metadata:
                self.__data_and_metadata.increment_data_ref_count()
        return initial_count+1

    def decrement_data_ref_count(self):
        with self.__data_ref_count_mutex:
            assert self.__data_ref_count > 0
            self.__data_ref_count -= 1
            final_count = self.__data_ref_count
            if final_count == 0 and self.__data_and_metadata:
                self.__data_and_metadata.decrement_data_ref_count()
        return final_count

    # used for testing
    @property
    def is_data_loaded(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_data_valid

    @property
    def has_data(self) -> bool:
        return self.__data_and_metadata is not None

    @property
    def data_shape(self):
        return self.__data_and_metadata.data_shape if self.__data_and_metadata else None

    @property
    def data_dtype(self):
        return self.__data_and_metadata.data_dtype if self.__data_and_metadata else None

    @property
    def is_data_1d(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_data_1d

    @property
    def is_data_2d(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_data_2d

    @property
    def is_data_3d(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_data_3d

    @property
    def is_data_4d(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_data_4d

    @property
    def is_data_rgb(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_data_rgb

    @property
    def is_data_rgba(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_data_rgba

    @property
    def is_datum_1d(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_datum_1d

    @property
    def is_datum_2d(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_datum_2d

    @property
    def is_data_rgb_type(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_data_rgb_type

    @property
    def is_data_scalar_type(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_data_scalar_type

    @property
    def is_data_complex_type(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_data_complex_type

    @property
    def is_data_bool(self) -> bool:
        return self.__data_and_metadata is not None and self.__data_and_metadata.is_data_bool

    def get_data_value(self, pos):
        return self.__data_and_metadata.get_data_value(pos) if self.__data_and_metadata else None

    @property
    def size_and_data_format_as_string(self) -> str:
        return self.__data_and_metadata.size_and_data_format_as_string if self.__data_and_metadata else _("No Data")


class SessionManager(abc.ABC):

    @property
    @abc.abstractmethod
    def current_session_id(self) -> str:
        pass


# dates are _local_ time and must use this specific ISO 8601 format. 2013-11-17T08:43:21.389391
# time zones are offsets (east of UTC) in the following format "+HHMM" or "-HHMM"
# daylight savings times are time offset (east of UTC) in format "+MM" or "-MM"
# time zone name is for display only and has no specified format

class LibraryItem(Observable.Observable, Persistence.PersistentObject):
    """
    Data items represent a data, description, display, and graphics within a library.

    * *created* a datetime item
    * *session_id* a string representing the session

    *Descriptive Metadata*

    * *title* a string (single line)
    * *caption* a string (multiple lines)
    * *rating* an integer star rating (0 to 5)
    * *flag* a flag (-1, 0, 1)

    *Modification Dates*

    Data items keep track of a created datetime, which is set once when the object is created; a modified datetime,
    which is updated whenever the object or its related objects get modified.

    Data items also provide a data modified datetime, which is the latest data modified datetime from all of the data
    sources.

    *Notifications*

    Data items will emit the following notifications to listeners. Listeners should take care to not call functions
    which result in cycles of notifications. For instance, functions handling data_item_content_changed should not use
    functions that will trigger the data to be computed which will emit data_item_content_changed, resulting in a cycle.

    * data_item_content_changed(data_item, changes)

    data_item_content_changed is invoked when the content of the data item changes. The changes parameter is a set of
    changes from DATA, METADATA, DISPLAYS, SOURCE. This may be called on a thread.

    *Miscellaneous*

    Transactions.

    Live state.

    Snapshots and deep copies.

    Properties.

    Metadata Note: Until data source is merged into data item, metadata access cannot be done in this base class;
    instead it is implemented in the derived classes. It is a regular field in the composite library item; but in
    the data item, it is stored in the data source. When data source is merged, it will become a regular field
    and can be moved into this class.
    """

    writer_version = 12

    def __init__(self, item_uuid=None):
        super().__init__()
        global writer_version
        self.uuid = item_uuid if item_uuid else self.uuid
        self.__container_weak_ref = None
        self.__in_transaction_state = False
        self.__is_live = False
        self.__pending_write = True
        self.__write_delay_modified_count = 0
        self.persistent_object_context = None
        self.__persistent_storage = None
        self.define_property("created", datetime.datetime.utcnow(), converter=DatetimeToStringConverter(), changed=self.__description_property_changed)
        # windows utcnow has a resolution of 1ms, this sleep can guarantee unique times for all created times during a particular test.
        # this is not my favorite solution since it limits library item creation to 1000/s but until I find a better solution, this is my compromise.
        time.sleep(0.001)
        self.define_property("description", dict(), hidden=True, changed=self.__property_changed)
        self.define_property("source_file_path", validate=self.__validate_source_file_path, changed=self.__property_changed)
        self.define_property("session_id", validate=self.__validate_session_id, changed=self.__session_id_changed)
        self.define_property("category", "persistent", changed=self.__property_changed)
        self.define_property("session_metadata", dict(), copy_on_read=True, changed=self.__property_changed)
        self.define_property("timezone", Utility.get_local_timezone(), changed=self.__timezone_property_changed)
        self.define_property("timezone_offset", Utility.TimezoneMinutesToStringConverter().convert(Utility.local_utcoffset_minutes()), changed=self.__timezone_property_changed)
        self.define_property("source_uuid", converter=Converter.UuidToStringConverter())
        self.define_relationship("displays", Display.display_factory, insert=self.__insert_display, remove=self.__remove_display)
        self.__session_manager = None
        self.__source = None
        self.library_item_changed_event = Event.Event()
        self.item_changed_event = Event.Event()  # equivalent to library_item_changed_event
        self.metadata_changed_event = Event.Event()  # see Metadata Note above
        self.__display_ref_count = 0
        self.__change_count = 0
        self.__change_count_lock = threading.RLock()
        self.__content_changed = False
        self.__suspendable_storage_cache = None
        self.r_var = None
        self.about_to_be_removed_event = Event.Event()
        self._about_to_be_removed = False
        self._closed = False

    def __str__(self):
        return "{0} {1} ({2}, {3})".format(self.__repr__(), (self.title if self.title else _("Untitled")), str(self.uuid), self.date_for_sorting_local_as_string)

    def __copy__(self):
        assert False

    def __deepcopy__(self, memo):
        library_item_copy = self.__class__()
        # metadata
        library_item_copy.description = self.description
        library_item_copy.session_metadata = self.session_metadata
        library_item_copy.created = self.created
        library_item_copy.timezone = self.timezone
        library_item_copy.timezone_offset = self.timezone_offset
        library_item_copy.session_id = self.session_id
        library_item_copy.source_file_path = self.source_file_path
        # displays
        for display in copy.copy(library_item_copy.displays):
            library_item_copy.remove_display(display)
        for display in self.displays:
            library_item_copy.add_display(copy.deepcopy(display))
        memo[id(self)] = library_item_copy
        return library_item_copy

    def close(self):
        for display in self.displays:
            display.close()
        # close the storage handler
        if self.persistent_object_context:
            if self.persistent_storage:
                self.persistent_storage.close()
            self.persistent_storage = None
            self.persistent_object_context = None
        assert self._about_to_be_removed
        assert not self._closed
        self._closed = True
        self.__container_weak_ref = None

    @property
    def container(self):
        return self.__container_weak_ref()

    def about_to_close(self):
        pass

    def about_to_be_inserted(self, container):
        assert self.__container_weak_ref is None
        self.__container_weak_ref = weakref.ref(container)

    def about_to_be_removed(self):
        # called before close and before item is removed from its container
        self.about_to_be_removed_event.fire()
        for display in self.displays:
            display.about_to_be_removed()
        assert not self._about_to_be_removed
        self._about_to_be_removed = True

    def insert_model_item(self, container, name, before_index, item):
        """Insert a model item. Let this item's container do it if possible; otherwise do it directly.

        Passing responsibility to this item's container allows the library to easily track dependencies.
        However, if this item isn't yet in the library hierarchy, then do the operation directly.
        """
        if self.__container_weak_ref:
            self.container.insert_model_item(container, name, before_index, item)
        else:
            container.insert_item(name, before_index, item)

    def remove_model_item(self, container, name, item, *, safe: bool=False) -> typing.Optional[typing.Sequence]:
        """Remove a model item. Let this item's container do it if possible; otherwise do it directly.

        Passing responsibility to this item's container allows the library to easily track dependencies.
        However, if this item isn't yet in the library hierarchy, then do the operation directly.
        """
        if self.__container_weak_ref:
            return self.container.remove_model_item(container, name, item, safe=safe)
        else:
            container.remove_item(name, item)
            return None

    @property
    def persistent_storage(self):
        return self.__persistent_storage

    @persistent_storage.setter
    def persistent_storage(self, persistent_storage):
        self._set_persistent_storage(persistent_storage)

    def _set_persistent_storage(self, persistent_storage):
        self.__persistent_storage = persistent_storage

    def clone(self) -> "LibraryItem":
        library_item = self.__class__()
        library_item.uuid = self.uuid
        for display in copy.copy(library_item.displays):
            library_item.remove_display(display)
        for display in self.displays:
            library_item.add_display(display.clone())
        return library_item

    def snapshot(self):
        """Return a new library item which is a copy of this one with any dynamic behavior made static."""
        libary_item = self.__class__()
        # metadata
        libary_item.description = self.description
        libary_item.session_metadata = self.session_metadata
        libary_item.created = self.created
        libary_item.timezone = self.timezone
        libary_item.timezone_offset = self.timezone_offset
        libary_item.session_id = self.session_id
        libary_item.source_file_path = self.source_file_path
        for display in copy.copy(libary_item.displays):
            libary_item.remove_display(display)
        for display in self.displays:
            libary_item.add_display(copy.deepcopy(display))
        return libary_item

    @property
    def data_items(self):
        return tuple()

    def connect_data_items(self, lookup_data_item):
        pass

    def set_storage_cache(self, storage_cache):
        self.__suspendable_storage_cache = Cache.SuspendableCache(storage_cache)
        for display in self.displays:
            display.set_storage_cache(self._suspendable_storage_cache)

    @property
    def _suspendable_storage_cache(self):
        return self.__suspendable_storage_cache

    @property
    def in_transaction_state(self) -> bool:
        return self.__in_transaction_state

    def _enter_write_delay_state_inner(self):
        pass

    def _finish_pending_write_inner(self):
        pass

    def __enter_write_delay_state(self):
        self.__write_delay_modified_count = self.modified_count
        self._enter_write_delay_state_inner()
        if self.persistent_object_context:
            if self.persistent_storage:
                self.persistent_storage.write_delayed = True

    def __exit_write_delay_state(self):
        if self.persistent_object_context:
            if self.persistent_storage:
                self.persistent_storage.write_delayed = False
            self._finish_pending_write()

    def _finish_write(self):
        pass

    def _finish_pending_write(self):
        if self.__pending_write:
            # write the uuid and version explicitly
            self.persistent_object_context.property_changed(self, "uuid", str(self.uuid))
            self.persistent_object_context.property_changed(self, "version", DataItem.writer_version)
            self._finish_write()
            self.__pending_write = False
        else:
            if self.modified_count > self.__write_delay_modified_count:
                self.persistent_storage.update_properties()
            self._finish_pending_write_inner()

    def _transaction_state_entered(self):
        self.__in_transaction_state = True
        # first enter the write delay state.
        self.__enter_write_delay_state()
        # suspend disk caching
        if self.__suspendable_storage_cache:
            self.__suspendable_storage_cache.suspend_cache()

    def _transaction_state_exited(self):
        self.__in_transaction_state = False
        # being in the transaction state has the side effect of delaying the cache too.
        # spill whatever was into the local cache into the persistent cache.
        if self.__suspendable_storage_cache:
            self.__suspendable_storage_cache.spill_cache()
        # exit the write delay state.
        self.__exit_write_delay_state()

    @property
    def source(self):
        return self.__source

    @source.setter
    def source(self, source):
        self.__source = source
        self.source_uuid = source.uuid if source else None

    def persistent_object_context_changed(self):
        # handle case where persistent object context is set on an item that is already under transaction.
        # this can occur during acquisition. any other cases?
        super().persistent_object_context_changed()

        if self.__in_transaction_state:
            self.__enter_write_delay_state()

        def register():
            if self.__source is not None:
                pass

        def source_registered(source):
            self.__source = source
            register()

        def unregistered(source=None):
            pass

        if self.persistent_object_context:
            self.persistent_object_context.subscribe(self.source_uuid, source_registered, unregistered)
        else:
            unregistered()

    def _test_get_file_path(self):
        return self.persistent_storage._storage_handler.reference if self.persistent_storage else None

    # access properties

    def read_from_dict(self, properties):
        for display in copy.copy(self.displays):
            self.remove_display(display)
        # when reading, handle changes specially. first, put everything into a change
        # block; then make sure that no change notifications actually occur. this makes
        # sure things like cached values are preserved after reading.
        with self.data_item_changes():
            self._read_from_dict_inner(properties)
            if self.created is None:  # invalid timestamp -- set property to now but don't trigger change
                timestamp = datetime.datetime.now()
                self._get_persistent_property("created").value = timestamp
            self.__content_changed = False
        self.__pending_write = False

    def _read_from_dict_inner(self, properties):
        super().read_from_dict(properties)

    @property
    def properties(self):
        """ Used for debugging. """
        if self.persistent_object_context:
            return self.persistent_object_context.get_properties(self)
        return dict()

    def __insert_display(self, name, before_index, display):
        display.about_to_be_inserted(self)
        display.title = self.displayed_title

    def __remove_display(self, name, index, display):
        display.about_to_be_removed()
        display.close()

    def add_display(self, display):
        """Add a display, but do it through the container, so dependencies can be tracked."""
        self.insert_model_item(self, "displays", self.item_count("displays"), display)
        if self.__display_ref_count > 0:
            display._become_master()

    def remove_display(self, display: Display.Display) -> typing.Optional[typing.Sequence]:
        """Remove display, but do it through the container, so dependencies can be tracked."""
        if self.__display_ref_count > 0:
            display._relinquish_master()
        return self.remove_model_item(self, "displays", display)

    @property
    def primary_display_specifier(self):
        if len(self.displays) > 0:
            return DisplaySpecifier(self, self.displays[0])
        return DisplaySpecifier()

    @property
    def is_live(self):
        """Return whether this library item represents live acquisition."""
        return self.__is_live

    def _enter_live_state(self):
        self.__is_live = True
        self._notify_library_item_content_changed()  # this will affect is_live, so notify

    def _exit_live_state(self):
        self.__is_live = False
        self._notify_library_item_content_changed()  # this will affect is_live, so notify

    def __validate_session_id(self, value):
        assert value is None or datetime.datetime.strptime(value, "%Y%m%d-%H%M%S")
        return value

    def __session_id_changed(self, name, value):
        self.__property_changed(name, value)

    def data_item_changes(self):
        # return a context manager to batch up a set of changes so that listeners
        # are only notified after the last change is complete.
        library_item = self
        class LibraryItemChangeContextManager:
            def __enter__(self):
                library_item._begin_library_item_changes()
                return self
            def __exit__(self, type, value, traceback):
                library_item._end_library_item_changes()
        return LibraryItemChangeContextManager()

    def _begin_library_item_changes(self):
        with self.__change_count_lock:
            self.__change_count += 1

    def _end_library_item_changes(self):
        with self.__change_count_lock:
            self.__change_count -= 1
            change_count = self.__change_count
            content_changed = self.__content_changed
        # if the change count is now zero, it means that we're ready to notify listeners. but only notify listeners if
        # there are actual changes to report.
        if change_count == 0 and content_changed:
            self.library_item_changed_event.fire()
            self.item_changed_event.fire()
            self._item_changed()

    def _item_changed(self):
        for display in self.displays:
            display._item_changed()

    def __validate_source_file_path(self, value):
        value = str(value) if value is not None else str()
        if value:
            value = os.path.normpath(value)
        return value

    def __description_property_changed(self, name, value):
        self.__property_changed(name, value)
        self.__notify_description_changed()

    def __notify_description_changed(self):
        self._notify_library_item_content_changed()
        self._description_changed()

    def _description_changed(self):
        for display in self.displays:
            display.title = self.displayed_title

    def __property_changed(self, name, value):
        self.notify_property_changed(name)

    def __timezone_property_changed(self, name, value):
        self._update_timezone()
        self.__property_changed(name, value)

    def _update_timezone(self):
        pass

    def set_r_value(self, r_var: str, *, notify_changed=True) -> None:
        """Used to signal changes to the ref var, which are kept in document controller. ugh."""
        self.r_var = r_var
        self._description_changed()
        if notify_changed:  # set to False to set the r-value at startup; avoid marking it as a change
            self.__notify_description_changed()

    def increment_display_ref_count(self, amount: int=1):
        """Increment display reference count to indicate this library item is currently displayed."""
        display_ref_count = self.__display_ref_count
        self.__display_ref_count += amount
        if display_ref_count == 0:
            for display in self.displays:
                display._become_master()

    def decrement_display_ref_count(self, amount: int=1):
        """Decrement display reference count to indicate this library item is no longer displayed."""
        assert not self._closed
        self.__display_ref_count -= amount
        if self.__display_ref_count == 0:
            for display in self.displays:
                display._relinquish_master()

    @property
    def _display_ref_count(self):
        return self.__display_ref_count

    @property
    def displayed_title(self):
        if self.r_var:
            return "{0} ({1})".format(self.title, self.r_var)
        else:
            return self.title

    # call this when the listeners need to be updated (via data_item_content_changed).
    # Calling this method will send the data_item_content_changed method to each listener by using the method
    # data_item_changes.
    def _notify_library_item_content_changed(self):
        with self.data_item_changes():
            with self.__change_count_lock:
                self.__content_changed = True

    # date times

    @property
    def date_for_sorting(self):
        return self.created

    @property
    def date_for_sorting_local_as_string(self):
        date_utc = self.date_for_sorting
        tz_minutes = Utility.local_utcoffset_minutes(date_utc)
        date_local = date_utc + datetime.timedelta(minutes=tz_minutes)
        return date_local.strftime("%c")

    @property
    def created_local_as_string(self):
        return self.created_local.strftime("%c")

    @property
    def created_local(self):
        created_utc = self.created
        tz_minutes = Utility.local_utcoffset_minutes(created_utc)
        return created_utc + datetime.timedelta(minutes=tz_minutes)

    # access description

    @property
    def description(self):
        return copy.deepcopy(self._get_persistent_property_value("description"))

    @description.setter
    def description(self, description):
        assert description is not None
        self._set_persistent_property_value("description", copy.deepcopy(description))

    def set_description(self, description):
        self.description = description

    @property
    def _session_manager(self) -> SessionManager:
        return self.__session_manager

    def set_session_manager(self, session_manager: SessionManager) -> None:
        self.__session_manager = session_manager

    # override from storage to watch for changes to this library item. notify observers.
    def notify_property_changed(self, key):
        super().notify_property_changed(key)
        self._notify_library_item_content_changed()

    # description

    @property
    def title(self):
        return self.description.get("title", UNTITLED_STR)

    @title.setter
    def title(self, value):
        description = self.description
        description["title"] = str(value) if value is not None else str()
        self.description = description
        self.__description_property_changed("title", value)

    @property
    def caption(self):
        return self.description.get("caption", str())

    @caption.setter
    def caption(self, value):
        description = self.description
        description["caption"] = str(value) if value is not None else str()
        self.description = description
        self.__description_property_changed("caption", value)

    @property
    def flag(self):
        return self.description.get("flag", 0)

    @flag.setter
    def flag(self, value):
        description = self.description
        description["flag"] = max(min(int(value), 1), -1)
        self.description = description
        self.__description_property_changed("flag", value)

    @property
    def rating(self):
        return self.description.get("rating", 0)

    @rating.setter
    def rating(self, value):
        description = self.description
        description["rating"] = min(max(int(value), 0), 5)
        self.description = description
        self.__description_property_changed("rating", value)

    @property
    def text_for_filter(self):
        return " ".join([self.displayed_title, self.caption])


class CompositeLibraryItem(LibraryItem):
    """Composite library item consists of references to other library items."""

    def __init__(self, item_uuid=None):
        super().__init__(item_uuid=item_uuid)
        self.__data_items = list()
        self.__data_item_about_to_be_removed_listeners = list()
        self.add_display(Display.Display())  # always have one display, for now
        self.define_property("data_item_uuids", list(), converter=UuidsToStringsConverter(), changed=self.__property_changed)
        self.define_property("metadata", dict(), hidden=True, changed=self.__metadata_property_changed)
        self.__metadata = dict()

    def close(self):
        self.__unlisten_about_to_be_removed()
        super().close()

    def about_to_close(self):
        self.__unlisten_about_to_be_removed()
        super().about_to_close()

    def about_to_be_removed(self):
        self.__unlisten_about_to_be_removed()
        super().about_to_be_removed()

    def __unlisten_about_to_be_removed(self):
        for listener in self.__data_item_about_to_be_removed_listeners:
            listener.close()
        self.__data_item_about_to_be_removed_listeners.clear()

    def read_from_dict(self, properties):
        super().read_from_dict(properties)
        self.__metadata = self._get_persistent_property_value("metadata", dict())

    @property
    def data_items(self):
        return tuple(self.__data_items)

    @property
    def metadata(self):
        return copy.deepcopy(self.__metadata)

    @metadata.setter
    def metadata(self, metadata):
        assert isinstance(metadata, dict)
        self.__metadata = copy.deepcopy(metadata)
        self._set_persistent_property_value("metadata", self.__metadata)

    def connect_data_items(self, lookup_data_item):
        super().connect_data_items(lookup_data_item)
        for data_item_uuid in self.data_item_uuids:
            data_item = lookup_data_item(data_item_uuid)
            # check whether the data item is already in data items; this covers the case where data items are added
            # to the composite library item before it is added to the document when it will be connected.
            if data_item and not data_item in self.__data_items:
                before_index = len(self.__data_items)
                self.__data_items.append(data_item)
                self.__data_item_about_to_be_removed_listeners.append(data_item.about_to_be_removed_event.listen(functools.partial(self.__data_item_about_to_be_removed, data_item)))
                self.notify_insert_item("data_items", data_item, before_index)

    def append_data_item(self, data_item):
        self.insert_data_item(len(self.__data_items), data_item)

    def insert_data_item(self, before_index, data_item):
        assert data_item not in self.__data_items
        self.__data_items.insert(before_index, data_item)
        self.__data_item_about_to_be_removed_listeners.append(data_item.about_to_be_removed_event.listen(functools.partial(self.__data_item_about_to_be_removed, data_item)))
        data_item_uuids = self.data_item_uuids
        data_item_uuids.insert(before_index, data_item.uuid)
        self.data_item_uuids = data_item_uuids
        self.notify_property_changed("data_item_uuids")
        self.notify_insert_item("data_items", data_item, before_index)
        data_item.increment_display_ref_count(self._display_ref_count)

    def remove_data_item(self, data_item):
        index = self.__data_items.index(data_item)
        self.notify_remove_item("data_items", data_item, index)
        self.__data_item_about_to_be_removed_listeners[index].close()
        del self.__data_item_about_to_be_removed_listeners[index]
        self.__data_items.remove(data_item)
        data_item_uuids = self.data_item_uuids
        data_item_uuids.remove(data_item.uuid)
        self.data_item_uuids = data_item_uuids
        self.notify_property_changed("data_item_uuids")
        data_item.decrement_display_ref_count(self._display_ref_count)

    def __data_item_about_to_be_removed(self, data_item):
        self.remove_data_item(data_item)

    def increment_display_ref_count(self, amount: int=1):
        super().increment_display_ref_count(amount)
        for data_item in self.data_items:
            data_item.increment_display_ref_count(amount)

    def decrement_display_ref_count(self, amount: int=1):
        super().decrement_display_ref_count(amount)
        for data_item in self.data_items:
            data_item.decrement_display_ref_count(amount)

    def __property_changed(self, name, value):
        self.notify_property_changed(name)

    def __metadata_property_changed(self, name, value):
        self.__property_changed(name, value)
        self.metadata_changed_event.fire()

    @property
    def size_and_data_format_as_string(self) -> str:
        return "{0} ({1})".format(_("Composite"), len(self.__data_items))


class DataItem(LibraryItem):

    def __init__(self, data=None, item_uuid=None, large_format=False):
        super().__init__(item_uuid)
        self.add_display(Display.Display())  # always have one display, for now
        self.large_format = large_format
        self.define_item("data_source", data_source_factory, item_changed=self.__data_source_changed)
        self.data_item_changed_event = Event.Event()  # anything has changed
        self.data_changed_event = Event.Event()  # data has changed
        self.__write_delay_data_changed = False
        self.__data_source_metadata_changed_event_listener = None
        self.__data_source_data_changed_event_listener = None
        self.__change_thread = None
        self.__change_count = 0
        self.__change_count_lock = threading.RLock()
        self.__change_changed = False
        self.__change_data_changed = False
        self.__pending_xdata_lock = threading.RLock()
        self.__pending_xdata = None
        if data is not None:
            self.set_data_source(BufferedDataSource(data))
        self.__update_displays()

    def __deepcopy__(self, memo):
        data_item_copy = super().__deepcopy__(memo)
        # format
        data_item_copy.large_format = self.large_format
        # data source
        data_item_copy.set_data_source(copy.deepcopy(self.data_source))
        data_item_copy.__update_displays()
        return data_item_copy

    def close(self):
        data_source = self.data_source
        if data_source:
            self.__disconnect_data_source(data_source)
            data_source.close()
        super().close()

    def about_to_be_removed(self):
        data_source = self.data_source
        if data_source:
            data_source.about_to_be_removed()
        super().about_to_be_removed()

    def _set_persistent_storage(self, persistent_storage):
        super()._set_persistent_storage(persistent_storage)
        if self.data_source:
            self.data_source.persistent_storage = persistent_storage

    def clone(self) -> "DataItem":
        data_item = super().clone()
        data_source = self.data_source
        if data_source:
            data_item.set_data_source(data_source.clone())
        data_item.__update_displays()
        return data_item

    def snapshot(self):
        data_item = super().snapshot()
        # format
        data_item.large_format = self.large_format
        # data sources
        data_source = self.data_source
        if data_source:
            data_item.set_data_source(data_source.snapshot())
        data_item.__update_displays()
        return data_item

    def _transaction_state_entered(self):
        super()._transaction_state_entered()
        # tell each data source to load its data.
        # this prevents paging in and out.
        data_source = self.data_source
        if data_source:
            data_source.increment_data_ref_count()

    def _transaction_state_exited(self):
        super()._transaction_state_exited()
        # tell each data source to unload its data.
        data_source = self.data_source
        if data_source:
            data_source.decrement_data_ref_count()

    def _read_from_dict_inner(self, properties):
        super()._read_from_dict_inner(properties)
        self.__update_displays()  # this ensures that the display will validate

    def finish_reading(self):
        super().finish_reading()

    def _finish_write(self):
        super()._finish_write()
        if self.data_source:
            self.persistent_storage.update_data(self.data)

    def _enter_write_delay_state_inner(self):
        super()._enter_write_delay_state_inner()
        self.__write_delay_data_changed = False

    def _finish_pending_write_inner(self):
        super()._finish_pending_write_inner()
        if self.__write_delay_data_changed:
            self.persistent_storage.update_data(self.data)

    @property
    def date_for_sorting(self):
        data_modified_list = list()
        data_source = self.data_source
        if data_source:
            data_modified = data_source.data_modified
            if data_modified:
                data_modified_list.append(data_modified)
            else:
                data_modified_list.append(self.created)
        if len(data_modified_list):
            return max(data_modified_list)
        return super().date_for_sorting

    def update_data_and_metadata(self, data_and_metadata: DataAndMetadata.DataAndMetadata) -> None:
        assert threading.current_thread() == threading.main_thread()
        with self.data_item_changes():
            self.set_xdata(data_and_metadata)
            self.timezone = Utility.get_local_timezone()
            self.timezone_offset = Utility.TimezoneMinutesToStringConverter().convert(Utility.local_utcoffset_minutes())

    def set_pending_xdata(self, xd: DataAndMetadata.DataAndMetadata) -> None:
        with self.__pending_xdata_lock:
            self.__pending_xdata = xd

    def update_to_pending_xdata(self):
        with self.__pending_xdata_lock:
            pending_xdata = self.__pending_xdata
            self.__pending_xdata = None
        if pending_xdata:
            self.update_data_and_metadata(pending_xdata)

    def __handle_data_changed(self, data_source):
        self.__change_changed = True
        self.__change_data_changed = True
        self.timezone = Utility.get_local_timezone()
        self.timezone_offset = Utility.TimezoneMinutesToStringConverter().convert(Utility.local_utcoffset_minutes())
        if self._session_manager:
            self.session_id = self._session_manager.current_session_id

    def __handle_metadata_changed(self):
        self.__change_changed = True
        self.metadata_changed_event.fire()

    def __data_source_changed(self, name, old_data_source, data_source):
        if old_data_source:
            old_data_source.about_to_be_removed()
            self.__disconnect_data_source(old_data_source)
            old_data_source.close()

        if data_source:
            data_source.persistent_storage = self.persistent_storage
            data_source.timezone = self.timezone
            data_source.timezone_offset = self.timezone_offset
            self.__data_source_data_changed_event_listener = data_source.data_changed_event.listen(self.__handle_data_changed)
            # being in transaction state means that data sources have their data loaded.
            # so load data here to keep the books straight when the transaction state is exited.
            if self.in_transaction_state:
                data_source.increment_data_ref_count()
            self.__data_source_metadata_changed_event_listener = data_source.metadata_changed_event.listen(self.__handle_metadata_changed)
            self._notify_library_item_content_changed()
            # the document model watches for new data sources via observing.
            # send this message to make data_source observable.
            self.notify_set_item("data_source", data_source)

    def __disconnect_data_source(self, data_source):
        if self.__data_source_metadata_changed_event_listener:
            self.__data_source_metadata_changed_event_listener.close()
            self.__data_source_metadata_changed_event_listener = None
        if self.__data_source_data_changed_event_listener:
            self.__data_source_data_changed_event_listener.close()
            self.__data_source_data_changed_event_listener = None
        # being in transaction state means that data sources have their data loaded.
        # so unload data here to keep the books straight when the transaction state is exited.
        if self.in_transaction_state:
            data_source.decrement_data_ref_count()
        data_source.persistent_storage = None
        # the document model watches for new data sources via observing.
        # send this message to make data_source observable.
        self.notify_clear_item("data_source")

    def set_data_source(self, data_source):
        """Set data source."""
        self.set_item("data_source", data_source)

    @property
    def data_source(self) -> typing.Optional[BufferedDataSource]:
        return self.get_item("data_source")

    def increment_display_ref_count(self, amount: int=1):
        super().increment_display_ref_count(amount)
        for _ in range(amount):
            self.increment_data_ref_count()

    def decrement_display_ref_count(self, amount: int=1):
        super().decrement_display_ref_count(amount)
        for _ in range(amount):
            self.decrement_data_ref_count()

    def increment_data_ref_count(self):
        data_source = self.data_source
        if data_source:
            data_source.increment_data_ref_count()

    def decrement_data_ref_count(self):
        data_source = self.data_source
        if data_source:
            data_source.decrement_data_ref_count()

    # primary display

    @property
    def primary_display_specifier(self):
        data_source = self.data_source
        if data_source:
            return DisplaySpecifier(self, self.displays[0])
        return DisplaySpecifier()

    def _update_timezone(self):
        with self.data_source_changes():
            data_source = self.data_source
            if data_source:
                data_source.timezone = self.timezone
                data_source.timezone_offset = self.timezone_offset
            super()._update_timezone()

    def ensure_data_source(self) -> None:
        if not self.data_source:
            self.set_data_source(BufferedDataSource())

    @property
    def data(self) -> numpy.ndarray:
        return self.__get_data()

    @property
    def xdata(self) -> DataAndMetadata.DataAndMetadata:
        return self.data_source.data_and_metadata if self.data_source else None

    def set_data(self, data: numpy.ndarray, data_modified: datetime.datetime=None) -> None:
        self.set_xdata(DataAndMetadata.new_data_and_metadata(data, data_modified))

    def set_xdata(self, xdata: DataAndMetadata.DataAndMetadata, data_modified: datetime.datetime=None) -> None:
        with self.data_source_changes():
            self.ensure_data_source()
            self.data_source.set_data_and_metadata(xdata, data_modified)

    # grab a data reference as a context manager. the object
    # returned defines data and data properties. reading data
    # should use the data property. writing data (if allowed) should
    # assign to the data property.
    def data_ref(self):
        get_data = self.__get_data
        set_data = self.__set_data
        class DataAccessor:
            def __init__(self, data_item):
                self.__data_item = data_item
            def __enter__(self):
                self.__data_item.increment_data_ref_count()
                return self
            def __exit__(self, type, value, traceback):
                self.__data_item.decrement_data_ref_count()
            @property
            def data(self):
                return get_data()
            @data.setter
            def data(self, value):
                set_data(value)
            def data_updated(self):
                set_data(get_data())
            @property
            def master_data(self):
                return get_data()
            @master_data.setter
            def master_data(self, value):
                set_data(value)
            def master_data_updated(self):
                set_data(get_data())
        return DataAccessor(self)

    def __get_data(self):
        return self.data_source._get_data() if self.data_source else None

    def __set_data(self, data, data_modified=None):
        with self.data_source_changes():
            self.data_source._update_data(data, data_modified)

    def data_source_changes(self):
        # return a context manager to batch up a set of changes so that listeners
        # are only notified after the last change is complete.
        begin_changes = self.__begin_changes
        end_changes = self.__end_changes
        class ChangeContextManager:
            def __enter__(self):
                begin_changes()
                return self
            def __exit__(self, type, value, traceback):
                end_changes()
        return ChangeContextManager()

    def __begin_changes(self):
        with self.__change_count_lock:
            if self.__change_count == 0:
                self.__change_thread = threading.current_thread()
            else:
                if self.__change_thread != threading.current_thread():
                    warnings.warn('begin changes from different threads', RuntimeWarning, stacklevel=2)
            self.__change_count += 1

    def __end_changes(self):
        changed = False
        data_changed = False
        with self.__change_count_lock:
            if self.__change_count == 1:
                self.__change_thread = None
            else:
                if self.__change_thread != threading.current_thread():
                    warnings.warn('end changes from different threads', RuntimeWarning, stacklevel=2)
            self.__change_count -= 1
            change_count = self.__change_count
            if change_count == 0:
                changed = self.__change_changed
                self.__change_changed = False
                data_changed = self.__change_data_changed
                self.__change_data_changed = False
        # if the change count is now zero, it means that we're ready
        # to pass on the next value.
        if change_count == 0:
            if data_changed:
                self.data_changed_event.fire()
            if changed:
                self.__update_displays()
                if not self._is_reading:
                    self.__write_delay_data_changed = True
                    self._notify_library_item_content_changed()
                    self.data_item_changed_event.fire()

    def __update_displays(self):
        data_and_metadata = self.xdata
        for display in self.displays:
            display.update_data(data_and_metadata)

    @property
    def data_modified(self) -> datetime.datetime:
        return self.data_source.data_modified if self.data_source else None

    @data_modified.setter
    def data_modified(self, value: datetime.datetime) -> None:
        with self.data_source_changes():
            if self.data_source:
                self.data_source.data_modified = value

    @property
    def intensity_calibration(self) -> Calibration.Calibration:
        return self.data_source.intensity_calibration if self.data_source else None

    @intensity_calibration.setter
    def intensity_calibration(self, intensity_calibration: Calibration.Calibration) -> None:
        with self.data_source_changes():
            if self.data_source:
                self.data_source.set_intensity_calibration(intensity_calibration)

    def set_intensity_calibration(self, intensity_calibration):
        self.intensity_calibration = intensity_calibration

    @property
    def dimensional_calibrations(self) -> typing.List[Calibration.Calibration]:
        return self.data_source.dimensional_calibrations if self.data_source else list()

    @dimensional_calibrations.setter
    def dimensional_calibrations(self, dimensional_calibrations: typing.Sequence[Calibration.Calibration]) -> None:
        with self.data_source_changes():
            if self.data_source:
                self.data_source.set_dimensional_calibrations(dimensional_calibrations)

    def set_dimensional_calibrations(self, dimensional_calibrations: typing.Sequence[Calibration.Calibration]) -> None:
        self.dimensional_calibrations = dimensional_calibrations

    def set_dimensional_calibration(self, dimension: int, calibration: Calibration.Calibration) -> None:
        dimensional_calibrations = self.dimensional_calibrations
        while len(dimensional_calibrations) <= dimension:
            dimensional_calibrations.append(Calibration.Calibration())
        dimensional_calibrations[dimension] = calibration
        self.set_dimensional_calibrations(dimensional_calibrations)

    @property
    def metadata(self) -> dict:
        return self.data_source.metadata if self.data_source else dict()

    @metadata.setter
    def metadata(self, value: dict) -> None:
        with self.data_source_changes():
            if self.data_source:
                self.data_source.metadata = value

    @property
    def has_data(self) -> bool:
        return self.data_source.has_data if self.data_source else False

    # used for testing
    @property
    def is_data_loaded(self) -> bool:
        return self.data_source.is_data_loaded if self.data_source else False

    @property
    def data_metadata(self):
        return self.data_source.data_metadata if self.data_source else None

    @property
    def data_shape(self):
        return self.data_source.data_shape if self.data_source else None

    @property
    def data_dtype(self):
        return self.data_source.data_dtype if self.data_source else None

    @property
    def dimensional_shape(self):
        return self.data_source.dimensional_shape if self.data_source else list()

    @property
    def datum_dimension_count(self) -> int:
        return self.data_source.datum_dimension_count if self.data_source else 0

    @property
    def collection_dimension_count(self) -> int:
        return self.data_source.collection_dimension_count if self.data_source else 0

    @property
    def is_collection(self):
        return self.data_source.is_collection if self.data_source else None

    @property
    def is_sequence(self) -> bool:
        return self.data_source.is_sequence if self.data_source else None

    @property
    def is_data_1d(self) -> bool:
        return self.data_source is not None and self.data_source.is_data_1d

    @property
    def is_data_2d(self) -> bool:
        return self.data_source is not None and self.data_source.is_data_2d

    @property
    def is_data_3d(self) -> bool:
        return self.data_source is not None and self.data_source.is_data_3d

    @property
    def is_data_4d(self) -> bool:
        return self.data_source is not None and self.data_source.is_data_4d

    @property
    def is_data_rgb(self) -> bool:
        return self.data_source is not None and self.data_source.is_data_rgb

    @property
    def is_data_rgba(self) -> bool:
        return self.data_source is not None and self.data_source.is_data_rgba

    @property
    def is_datum_1d(self) -> bool:
        return self.data_source is not None and self.data_source.is_datum_1d

    @property
    def is_datum_2d(self) -> bool:
        return self.data_source is not None and self.data_source.is_datum_2d

    @property
    def is_data_rgb_type(self) -> bool:
        return self.data_source is not None and self.data_source.is_data_rgb_type

    @property
    def is_data_scalar_type(self) -> bool:
        return self.data_source is not None and self.data_source.is_data_scalar_type

    @property
    def is_data_complex_type(self) -> bool:
        return self.data_source is not None and self.data_source.is_data_complex_type

    @property
    def is_data_bool(self) -> bool:
        return self.data_source is not None and self.data_source.is_data_bool

    def get_data_value(self, pos):
        return self.data_source.get_data_value(pos) if self.data_source else None

    @property
    def size_and_data_format_as_string(self) -> str:
        return self.data_source.size_and_data_format_as_string if self.data_source else _("No Data")

    def has_metadata_value(self, key: str) -> bool:
        return Metadata.has_metadata_value(self, key)

    def get_metadata_value(self, key: str) -> typing.Any:
        return Metadata.get_metadata_value(self, key)

    def set_metadata_value(self, key: str, value: typing.Any) -> None:
        Metadata.set_metadata_value(self, key, value)

    def delete_metadata_value(self, key: str) -> None:
        Metadata.delete_metadata_value(self, key)


class DisplaySpecifier:
    """Specify a Display contained within a DataItem."""

    def __init__(self, library_item: LibraryItem=None, display: Display.Display=None):
        self.library_item = library_item
        self.display = display

    def __eq__(self, other):
        return self.library_item == other.library_item and self.display == other.display

    def __ne__(self, other):
        return self.library_item != other.library_item or self.display != other.display

    @property
    def data_item(self):
        return self.library_item if isinstance(self.library_item, DataItem) else None

    @property
    def composite_library_item(self):
        return self.library_item if isinstance(self.library_item, CompositeLibraryItem) else None

    @classmethod
    def from_display(cls, display):
        library_item = display.container if display else None
        return cls(library_item, display)

    @classmethod
    def from_library_item(cls, library_item):
        display = library_item.displays[0] if library_item and len(library_item.displays) > 0 else None
        return cls(library_item, display)

    @classmethod
    def from_data_item(cls, data_item):
        display = data_item.displays[0] if data_item and len(data_item.displays) > 0 else None
        return cls(data_item, display)


def sort_by_date_key(data_item):
    """ A sort key to for the created field of a data item. The sort by uuid makes it determinate. """
    return data_item.title + str(data_item.uuid) if data_item.is_live else str(), data_item.date_for_sorting, str(data_item.uuid)


def new_data_item(data_and_metadata: DataAndMetadata.DataAndMetadata=None) -> DataItem:
    data_item = DataItem(large_format=data_and_metadata and len(data_and_metadata.dimensional_shape) > 2)
    data_item.ensure_data_source()
    data_item.set_xdata(data_and_metadata)
    return data_item


class DataSource:
    DATA_SOURCE_MIME_TYPE = "text/vnd.nion.display_source_type"

    def __init__(self, data_item: DataItem, graphic, changed_event):
        self.__data_item = data_item
        self.__graphic = graphic
        self.__changed_event = changed_event  # not public since it is passed in
        display = data_item.displays[0] if len(data_item.displays) > 0 else None
        self.__data_item_changed_event_listener = data_item.data_item_changed_event.listen(self.__changed_event.fire)
        self.__display_values_event_listener = display.display_data_will_change_event.listen(self.__changed_event.fire) if display else None
        self.__property_changed_listener = None
        def property_changed(key):
            self.__changed_event.fire()
        if graphic:
            self.__property_changed_listener = graphic.property_changed_event.listen(property_changed)
        def filter_property_changed(key):
            self.__changed_event.fire()
        self.__graphic_property_changed_listeners = list()
        def graphic_inserted(key, value, before_index):
            if key == "graphics":
                property_changed_listener = None
                if isinstance(graphic, (Graphics.SpotGraphic, Graphics.WedgeGraphic, Graphics.RingGraphic)):
                    property_changed_listener = value.property_changed_event.listen(filter_property_changed)
                    self.__changed_event.fire()
                self.__graphic_property_changed_listeners.insert(before_index, property_changed_listener)
        def graphic_removed(key, value, index):
            if key == "graphics":
                property_changed_listener = self.__graphic_property_changed_listeners.pop(index)
                if property_changed_listener:
                    property_changed_listener.close()
                    self.__changed_event.fire()
        self.__graphic_inserted_event_listener = display.item_inserted_event.listen(graphic_inserted) if display else None
        self.__graphic_removed_event_listener = display.item_removed_event.listen(graphic_removed) if display else None
        for graphic in display.graphics if display else list():
            property_changed_listener = None
            if isinstance(graphic, (Graphics.SpotGraphic, Graphics.WedgeGraphic, Graphics.RingGraphic)):
                property_changed_listener = graphic.property_changed_event.listen(filter_property_changed)
            self.__graphic_property_changed_listeners.append(property_changed_listener)

    def close(self):
        for graphic_property_changed_listener in self.__graphic_property_changed_listeners:
            if graphic_property_changed_listener:
                graphic_property_changed_listener.close()
        self.__graphic_property_changed_listeners = list()
        if self.__graphic_inserted_event_listener:
            self.__graphic_inserted_event_listener.close()
            self.__graphic_inserted_event_listener = None
        if self.__graphic_removed_event_listener:
            self.__graphic_removed_event_listener.close()
            self.__graphic_removed_event_listener = None
        if self.__property_changed_listener:
            self.__property_changed_listener.close()
            self.__property_changed_listener = None
        self.__data_item_changed_event_listener.close()
        self.__data_item_changed_event_listener = None
        if self.__display_values_event_listener:
            self.__display_values_event_listener.close()
            self.__display_values_event_listener = None

    @property
    def data_item(self):
        return self.__data_item

    @property
    def graphic(self):
        return self.__graphic

    @property
    def data(self) -> numpy.ndarray:
        return self.__data_item.data

    @property
    def xdata(self) -> typing.Optional[DataAndMetadata.DataAndMetadata]:
        return self.__data_item.xdata

    @property
    def display_xdata(self) -> DataAndMetadata.DataAndMetadata:
        return self.__data_item.displays[0].get_calculated_display_values(True).display_data_and_metadata

    @property
    def cropped_display_xdata(self) -> typing.Optional[DataAndMetadata.DataAndMetadata]:
        data_item = self.__data_item
        if data_item:
            displayed_xdata = self.display_xdata
            graphic = self.__graphic
            if graphic:
                if hasattr(graphic, "bounds") and displayed_xdata.is_data_2d:
                    return Core.function_crop(displayed_xdata, graphic.bounds)
                if hasattr(graphic, "interval") and displayed_xdata.is_data_1d:
                    return Core.function_crop_interval(displayed_xdata, graphic.interval)
            return displayed_xdata
        return None

    @property
    def cropped_xdata(self) -> typing.Optional[DataAndMetadata.DataAndMetadata]:
        data_item = self.__data_item
        if data_item:
            xdata = self.xdata
            graphic = self.__graphic
            if graphic:
                if hasattr(graphic, "bounds"):
                    return Core.function_crop(xdata, graphic.bounds)
                if hasattr(graphic, "interval"):
                    return Core.function_crop_interval(xdata, graphic.interval)
            return xdata
        return None

    @property
    def filtered_xdata(self) -> typing.Optional[DataAndMetadata.DataAndMetadata]:
        xdata = self.xdata
        if xdata.is_data_2d and xdata.is_data_complex_type:
            shape = xdata.data_shape
            mask = numpy.zeros(shape)
            for graphic in self.__data_item.displays[0].graphics:
                if isinstance(graphic, (Graphics.SpotGraphic, Graphics.WedgeGraphic, Graphics.RingGraphic)):
                    mask = numpy.logical_or(mask, graphic.get_mask(shape))
            return Core.function_fourier_mask(xdata, DataAndMetadata.DataAndMetadata.from_data(mask))
        return xdata

    @property
    def filter_xdata(self) -> typing.Optional[DataAndMetadata.DataAndMetadata]:
        shape = self.display_xdata.data_shape
        mask = numpy.zeros(shape)
        for graphic in self.__data_item.displays[0].graphics:
            if isinstance(graphic, (Graphics.SpotGraphic, Graphics.WedgeGraphic, Graphics.RingGraphic)):
                mask = numpy.logical_or(mask, graphic.get_mask(shape))
        return DataAndMetadata.DataAndMetadata.from_data(mask)



# standard libraries
import asyncio
import concurrent.futures
import copy
import datetime
import gettext
import json
import logging
import os
import shutil
import sys
import typing

# third party libraries
# None

# local libraries
from nion.swift import DataPanel
from nion.swift import DocumentController
from nion.swift import FilterPanel
from nion.swift import HistogramPanel
from nion.swift import InfoPanel
from nion.swift import Inspector
from nion.swift import MetadataPanel
from nion.swift import Panel
from nion.swift import SessionPanel
from nion.swift import Task
from nion.swift import Test
from nion.swift import ToolbarPanel
from nion.swift import Workspace
from nion.swift.model import Cache
from nion.swift.model import DataItem
from nion.swift.model import DocumentModel
from nion.swift.model import HardwareSource
from nion.swift.model import PlugInManager
from nion.swift.model import Utility
from nion.ui import Dialog
from nion.ui import Widgets
from nion.utils import Selection

_ = gettext.gettext

app = None


# facilitate bootstrapping the application
class Application:

    def __init__(self, ui, set_global=True, resources_path=None):
        global app

        ui.set_application_info("Nion Swift", "Nion", "nion.com")

        self.ui = ui
        self.ui.persistence_root = "3"  # sets of preferences
        self.__resources_path = resources_path
        self.version_str = "0.13.0"
        self.workspace_dir = None

        self.__event_loop = None

        if True or set_global:
            app = self  # hack to get the single instance set. hmm. better way?

        self.__document_model = None

        logger = logging.getLogger()
        logger.setLevel(logging.DEBUG)
        logger.addHandler(logging.StreamHandler())

        # a list of document controllers in the application.
        self.__document_controllers = []
        self.__menu_handlers = []

        # map these document controller events to listener tokens.
        # when the document controller closes, remove its listeners and
        # then remove it from the list of document controllers.
        # when the document controller requests a new document controller,
        # respond in this class by creating a new document controller.
        self.__did_close_event_listeners = dict()
        self.__create_new_event_listeners = dict()

        workspace_manager = Workspace.WorkspaceManager()
        workspace_manager.register_panel(SessionPanel.SessionPanel, "session-panel", _("Session"), ["left", "right"], "right", {"min-width": 320, "height": 80})
        workspace_manager.register_panel(DataPanel.DataPanel, "data-panel", _("Data Panel"), ["left", "right"], "left", {"min-width": 320, "height": 400})
        workspace_manager.register_panel(HistogramPanel.HistogramPanel, "histogram-panel", _("Histogram"), ["left", "right"], "right", {"min-width": 320, "height": 140})
        workspace_manager.register_panel(InfoPanel.InfoPanel, "info-panel", _("Info"), ["left", "right"], "right", {"min-width": 320, "height": 60})
        workspace_manager.register_panel(Inspector.InspectorPanel, "inspector-panel", _("Inspector"), ["left", "right"], "right", {"min-width": 320})
        workspace_manager.register_panel(Task.TaskPanel, "task-panel", _("Task Panel"), ["left", "right"], "right", {"min-width": 320})
        workspace_manager.register_panel(Panel.OutputPanel, "output-panel", _("Output"), ["bottom"], "bottom")
        workspace_manager.register_panel(ToolbarPanel.ToolbarPanel, "toolbar-panel", _("Toolbar"), ["top"], "top", {"height": 30})
        workspace_manager.register_panel(MetadataPanel.MetadataPanel, "metadata-panel", _("Metadata"), ["left", "right"], "right", {"width": 320, "height": 8})
        workspace_manager.register_filter_panel(FilterPanel.FilterPanel)

    def initialize(self, *, load_plug_ins=True, use_root_dir=True):
        # configure the event loop object
        logger = logging.getLogger()
        old_level = logger.level
        logger.setLevel(logging.INFO)
        self.__event_loop = asyncio.new_event_loop()  # outputs a debugger message!
        logger.setLevel(old_level)
        # load plug-ins
        if load_plug_ins:
            PlugInManager.load_plug_ins(self, get_root_dir() if use_root_dir else None)

    def deinitialize(self):
        # shut down hardware source manager, unload plug-ins, and really exit ui
        HardwareSource.HardwareSourceManager().close()
        PlugInManager.unload_plug_ins()
        with open(os.path.join(self.ui.get_data_location(), "PythonConfig.ini"), 'w') as f:
            f.write(sys.prefix + '\n')
        # give cancelled tasks a chance to finish
        self.__event_loop.stop()
        self.__event_loop.run_forever()
        try:
            # this assumes that all outstanding tasks finish in a reasonable time (i.e. no infinite loops).
            self.__event_loop.run_until_complete(asyncio.gather(*asyncio.Task.all_tasks(loop=self.__event_loop), loop=self.__event_loop))
        except concurrent.futures.CancelledError:
            pass
        # now close
        # due to a bug in Python libraries, the default executor needs to be shutdown explicitly before the event loop
        # see http://bugs.python.org/issue28464
        if self.__event_loop._default_executor:
            self.__event_loop._default_executor.shutdown()
        self.__event_loop.close()
        self.__event_loop = None
        self.ui.close()

    def run(self):
        """Alternate start which allows ui to control event loop."""
        self.ui.run(self)

    def exit(self):
        # close all document windows
        for document_controller in copy.copy(self.__document_controllers):
            # closing the document window will trigger the about_to_close event to be called which
            # will then call document controller close which will fire its did_close_event which will
            # remove the document controller from the list of document controllers.
            document_controller.request_close()
        # document model is reference counted; when the no document controller holds a reference to the
        # document model, it will be closed.

    def periodic(self) -> None:
        if self.__event_loop:  # special for shutdown
            self.__event_loop.stop()
            self.__event_loop.run_forever()

    @property
    def event_loop(self) -> asyncio.AbstractEventLoop:
        return self.__event_loop

    @property
    def document_model(self):
        return self.__document_model

    # for testing
    def _set_document_model(self, document_model):
        self.__document_model = document_model

    def __migrate_library(self, workspace_dir: str, welcome_message_enabled: bool=True) -> str:
        """ Migrate library to latest version. """
        library_path_11 = os.path.join(workspace_dir, "Nion Swift Workspace.nslib")
        library_path_12 = os.path.join(workspace_dir, "Nion Swift Library 12.nslib")
        if not os.path.exists(library_path_12):
            if os.path.exists(library_path_11):
                if welcome_message_enabled:
                    logging.debug("Migrating from library 11 to 12: %s -> %s", library_path_11, library_path_12)
                shutil.copyfile(library_path_11, library_path_12)
        return library_path_12

    def start(self, skip_choose=False, fixed_workspace_dir=None):
        """
            Start the application.

            Looks for workspace_location persistent string. If it doesn't find it, uses a default
            workspace location.

            Then checks to see if that workspace exists. If not and if skip_choose has not been
            set to True, asks the user for a workspace location. User may choose new folder or
            existing location. This works by putting up the dialog which will either call start
            again or exit.

            Creates workspace in location if it doesn't exist.

            Migrates database to latest version.

            Creates document model, resources path, etc.
        """
        if fixed_workspace_dir:
            workspace_dir = fixed_workspace_dir
        else:
            documents_dir = self.ui.get_document_location()
            workspace_dir = os.path.join(documents_dir, "Nion Swift Libraries")
            workspace_dir = self.ui.get_persistent_string("workspace_location", workspace_dir)
        cache_filename = "Nion Swift Cache {version}.nscache".format(version=DataItem.DataItem.writer_version)
        welcome_message_enabled = fixed_workspace_dir is None
        library_path = self.__migrate_library(workspace_dir, welcome_message_enabled)
        cache_path = os.path.join(workspace_dir, cache_filename)
        if not skip_choose and not os.path.exists(library_path):
            self.choose_library()
            return True
        self.workspace_dir = workspace_dir
        create_new_document = not os.path.exists(library_path)
        if create_new_document:
            if welcome_message_enabled:
                logging.debug("Creating new document: %s", library_path)
        else:
            if welcome_message_enabled:
                logging.debug("Using existing document %s", library_path)
        library_storage = DocumentModel.FilePersistentStorage(library_path)
        file_persistent_storage_system = DocumentModel.FileStorageSystem([os.path.join(workspace_dir, "Nion Swift Data {version}".format(version=DataItem.DataItem.writer_version))])
        counts = DocumentModel.read_data_items_version_stats(file_persistent_storage_system)
        if counts[2] > 0:

            assert fixed_workspace_dir is None

            def do_ignore():
                self.continue_start(cache_path, create_new_document, file_persistent_storage_system, library_storage, workspace_dir, True)
                return True

            def do_upgrade():
                self.continue_start(cache_path, create_new_document, file_persistent_storage_system, library_storage, workspace_dir, False)
                return True

            class UpgradeDialog(Dialog.ActionDialog):
                def __init__(self, ui):
                    super().__init__(ui)

                    self.add_button(_("Upgrade"), do_upgrade)

                    self.add_button(_("Ignore"), do_ignore)

                    column = self.ui.create_column_widget()
                    column.add_spacing(12)

                    row_one = self.ui.create_row_widget()
                    row_one.add_spacing(13)
                    row_one.add(self.ui.create_label_widget("{0} data items need to be updated.".format(counts[2])))
                    row_one.add_stretch()
                    row_one.add_spacing(13)
                    column.add(row_one)
                    column.add_spacing(4)

                    row_two = self.ui.create_row_widget()
                    row_two.add_spacing(13)
                    row_two.add(self.ui.create_label_widget("{0} data items are newer than this version and won't be loaded.".format(counts[1])))
                    row_two.add_stretch()
                    row_two.add_spacing(13)
                    column.add(row_two)
                    column.add_spacing(4)

                    row_three = self.ui.create_row_widget()
                    row_three.add_spacing(13)
                    row_three.add(self.ui.create_label_widget("{0} data items match this version.".format(counts[0])))
                    row_three.add_stretch()
                    row_three.add_spacing(13)
                    column.add(row_three)
                    column.add_spacing(4)

                    row_three = self.ui.create_row_widget()
                    row_three.add_spacing(13)
                    row_three.add(self.ui.create_label_widget("If you choose to upgrade, the upgraded items will not\nbe able to be loaded in earlier versions.".format(counts[0])))
                    row_three.add_stretch()
                    row_three.add_spacing(13)
                    column.add_spacing(8)
                    column.add(row_three)
                    column.add_spacing(4)

                    column.add_spacing(12)
                    column.add_stretch()

                    self.content.add(column)

            upgrade_dialog = UpgradeDialog(self.ui)
            upgrade_dialog.show()

        else:
            self.continue_start(cache_path, create_new_document, file_persistent_storage_system, library_storage, workspace_dir, True, welcome_message=welcome_message_enabled)

        return True

    def continue_start(self, cache_path, create_new_document, file_persistent_storage_system, library_storage, workspace_dir, ignore_older_files, welcome_message=True):
        storage_cache = Cache.DbStorageCache(cache_path)
        DocumentModel.DocumentModel.computation_min_period = 0.1
        auto_migrations = list()
        auto_migrations.append(DocumentModel.AutoMigration([os.path.join(workspace_dir, "Nion Swift Data")]))
        auto_migrations.append(DocumentModel.AutoMigration([os.path.join(workspace_dir, "Nion Swift Data 10")]))
        auto_migrations.append(DocumentModel.AutoMigration([os.path.join(workspace_dir, "Nion Swift Data 11")]))
        document_model = DocumentModel.DocumentModel(library_storage=library_storage,
                                                     persistent_storage_system=file_persistent_storage_system,
                                                     storage_cache=storage_cache, ignore_older_files=ignore_older_files,
                                                     auto_migrations=auto_migrations)
        document_model.create_default_data_groups()
        document_model.start_dispatcher()
        # parse the hardware aliases file
        alias_path = os.path.join(self.workspace_dir, "aliases.ini")
        HardwareSource.parse_hardware_aliases_config_file(alias_path)
        # create the document controller
        document_controller = self.create_document_controller(document_model, "library")
        if self.__resources_path is not None:
            document_model.create_sample_images(self.__resources_path)
        workspace_history = self.ui.get_persistent_object("workspace_history", list())
        if workspace_dir in workspace_history:
            workspace_history.remove(workspace_dir)
        workspace_history.insert(0, workspace_dir)
        self.ui.set_persistent_object("workspace_history", workspace_history)
        self.ui.set_persistent_string("workspace_location", workspace_dir)
        if welcome_message:
            logging.info("Welcome to Nion Swift.")
        if create_new_document and len(document_model.data_items) > 0:
            document_controller.selected_display_panel.set_display_panel_data_item(document_model.data_items[0])
            document_controller.selected_display_panel.perform_action("set_fill_mode")

    def stop(self):
        # program is really stopping, clean up.
        self.deinitialize()

    def get_recent_workspace_file_paths(self):
        workspace_history = self.ui.get_persistent_object("workspace_history", list())
        # workspace_history = ["/Users/cmeyer/Movies/Crap/Test1", "/Users/cmeyer/Movies/Crap/Test7_new"]
        return [file_path for file_path in workspace_history if file_path != self.workspace_dir and os.path.exists(file_path)]

    def switch_library(self, recent_workspace_file_path, skip_choose=False, fixed_workspace_dir=None):
        self.exit()
        self.ui.set_persistent_string("workspace_location", recent_workspace_file_path)
        self.start(skip_choose=skip_choose, fixed_workspace_dir=fixed_workspace_dir)

    def clear_libraries(self):
        self.ui.remove_persistent_key("workspace_history")

    def pose_open_library_dialog(self) -> typing.Optional[str]:
        documents_dir = self.ui.get_document_location()
        workspace_dir, directory = self.ui.get_existing_directory_dialog(_("Choose Library Folder"), documents_dir)
        if workspace_dir:
            path = os.path.join(workspace_dir, "Nion Swift Workspace.nslib")
            if not os.path.exists(path):
                with open(path, "w") as fp:
                    json.dump({}, fp)
            return workspace_dir
        return None

    def open_library(self):
        workspace_dir = self.pose_open_library_dialog()
        if workspace_dir:
            app.switch_library(workspace_dir)

    def new_library(self):

        class NewLibraryDialog(Dialog.ActionDialog):

            def __init__(self, ui, app):
                super().__init__(ui, title=_("New Library"), app=app, persistent_id="new_library_dialog")

                self.directory = self.ui.get_persistent_string("library_directory", self.ui.get_document_location())

                library_base_name = _("Nion Swift Library") + " " + datetime.datetime.now().strftime("%Y%m%d")
                library_base_index = 0
                library_base_index_str = ""
                while os.path.exists(os.path.join(self.directory, library_base_name + library_base_index_str)):
                    library_base_index += 1
                    library_base_index_str = " " + str(library_base_index)

                self.library_name = library_base_name + library_base_index_str

                def handle_new():
                    self.library_name = self.__library_name_field.text
                    workspace_dir = os.path.join(self.directory, self.library_name)
                    Cache.db_make_directory_if_needed(workspace_dir)
                    path = os.path.join(workspace_dir, "Nion Swift Workspace.nslib")
                    if not os.path.exists(path):
                        with open(path, "w") as fp:
                            json.dump({}, fp)
                    if os.path.exists(path):
                        app.switch_library(workspace_dir)
                        return True
                    return False

                def handle_new_and_close():
                    handle_new()
                    self.request_close()
                    return False

                column = self.ui.create_column_widget()

                directory_header_row = self.ui.create_row_widget()
                directory_header_row.add_spacing(13)
                directory_header_row.add(self.ui.create_label_widget(_("Libraries Folder: "), properties={"font": "bold"}))
                directory_header_row.add_stretch()
                directory_header_row.add_spacing(13)

                show_directory_row = self.ui.create_row_widget()
                show_directory_row.add_spacing(26)
                directory_label = self.ui.create_label_widget(self.directory)
                show_directory_row.add(directory_label)
                show_directory_row.add_stretch()
                show_directory_row.add_spacing(13)

                choose_directory_row = self.ui.create_row_widget()
                choose_directory_row.add_spacing(26)
                choose_directory_button = self.ui.create_push_button_widget(_("Choose..."))
                choose_directory_row.add(choose_directory_button)
                choose_directory_row.add_stretch()
                choose_directory_row.add_spacing(13)

                library_name_header_row = self.ui.create_row_widget()
                library_name_header_row.add_spacing(13)
                library_name_header_row.add(self.ui.create_label_widget(_("Library Name: "), properties={"font": "bold"}))
                library_name_header_row.add_stretch()
                library_name_header_row.add_spacing(13)

                library_name_row = self.ui.create_row_widget()
                library_name_row.add_spacing(26)
                library_name_field = self.ui.create_line_edit_widget(properties={"width": 400})
                library_name_field.text = self.library_name
                library_name_field.on_return_pressed = handle_new_and_close
                library_name_field.on_escape_pressed = self.request_close
                library_name_row.add(library_name_field)
                library_name_row.add_stretch()
                library_name_row.add_spacing(13)

                column.add_spacing(12)
                column.add(directory_header_row)
                column.add_spacing(8)
                column.add(show_directory_row)
                column.add_spacing(8)
                column.add(choose_directory_row)
                column.add_spacing(16)
                column.add(library_name_header_row)
                column.add_spacing(8)
                column.add(library_name_row)
                column.add_stretch()
                column.add_spacing(16)

                def choose() -> None:
                    existing_directory, directory = self.ui.get_existing_directory_dialog(_("Choose Library Directory"), self.directory)
                    if existing_directory:
                        self.directory = existing_directory
                        directory_label.text = self.directory
                        self.ui.set_persistent_string("library_directory", self.directory)

                choose_directory_button.on_clicked = choose

                self.add_button(_("Cancel"), lambda: True)
                self.add_button(_("Create Library"), handle_new)

                self.content.add(column)

                self.__library_name_field = library_name_field

            def show(self):
                super().show()
                self.__library_name_field.focused = True

        new_library_dialog = NewLibraryDialog(self.ui, self)
        new_library_dialog.show()

    def choose_library(self):

        pose_open_library_dialog_fn = self.pose_open_library_dialog

        workspace_history = self.ui.get_persistent_object("workspace_history", list())
        nslib_paths = [os.path.join(file_path, "Nion Swift Workspace.nslib") for file_path in workspace_history]
        items = [(path, datetime.datetime.fromtimestamp(os.path.getmtime(path))) for path in nslib_paths if os.path.exists(path)]

        class ChooseLibraryDialog(Dialog.ActionDialog):

            def __init__(self, ui, app):
                super().__init__(ui, _("Choose Library"))

                current_item_ref = [None]

                def handle_choose():
                    current_item = current_item_ref[0]
                    if current_item:
                        app.switch_library(current_item)
                        return True
                    return False

                def handle_new():
                    workspace_dir = pose_open_library_dialog_fn()
                    if workspace_dir:
                        items.insert(0, (workspace_dir, datetime.datetime.now()))
                        list_widget.items = items
                        list_widget.set_selected_index(0)
                        app.switch_library(workspace_dir)
                        return True
                    return False

                self.add_button(_("New..."), handle_new)
                self.add_button(_("Other..."), handle_new)
                self.add_button(_("Cancel"), lambda: True)
                self.add_button(_("Choose"), handle_choose)

                path_label = ui.create_label_widget()

                prompt_row = ui.create_row_widget()
                prompt_row.add_spacing(13)
                prompt_row.add(ui.create_label_widget(_("Which library do you want Nion Swift to use?"), properties={"stylesheet": "font-weight: bold"}))
                prompt_row.add_spacing(13)
                prompt_row.add_stretch()

                explanation1_row = ui.create_row_widget()
                explanation1_row.add_spacing(13)
                explanation1_row.add(ui.create_label_widget(_("You can select a library from the list, find another library, or create a new library.")))
                explanation1_row.add_spacing(13)
                explanation1_row.add_stretch()

                explanation2_row = ui.create_row_widget()
                explanation2_row.add_spacing(13)
                explanation2_row.add(ui.create_label_widget(_("The same library will be used the next time you open Nion Swift.")))
                explanation2_row.add_spacing(13)
                explanation2_row.add_stretch()

                def selection_changed(indexes):
                    if len(indexes) == 1:
                        item = items[list(indexes)[0]]
                        current_item_ref[0] = os.path.dirname(item[0])
                        path_label.text = os.path.dirname(item[0])
                    else:
                        current_item_ref[0] = None
                        path_label.text = None

                def stringify_item(item):
                    date_utc = item[1]
                    tz_minutes = Utility.local_utcoffset_minutes(date_utc)
                    date_local = date_utc + datetime.timedelta(minutes=tz_minutes)
                    return str(os.path.basename(os.path.dirname(item[0]))) + " (" + date_local.strftime("%c") + ")"

                def item_selected(index):
                    item = items[index]
                    current_item_ref[0] = os.path.dirname(item[0])
                    path_label.text = os.path.dirname(item[0])
                    handle_choose()
                    self.request_close()

                list_widget = Widgets.StringListWidget(ui, items, Selection.Style.single_or_none, stringify_item, properties={"min-height": 200, "min-width": 560})
                list_widget.on_selection_changed = selection_changed
                list_widget.on_item_selected = item_selected
                list_widget.on_cancel = self.request_close
                if len(items) > 0:
                    list_widget.set_selected_index(0)

                items_row = ui.create_row_widget()
                items_row.add_spacing(13)
                items_row.add(list_widget)
                items_row.add_spacing(13)
                items_row.add_stretch()

                path_row = ui.create_row_widget()
                path_row.add_spacing(13)
                path_row.add(path_label)
                path_row.add_spacing(13)
                path_row.add_stretch()

                column = ui.create_column_widget()
                column.add_spacing(18)
                column.add(prompt_row)
                column.add_spacing(6)
                column.add(explanation1_row)
                column.add(explanation2_row)
                column.add_spacing(12)
                column.add(items_row)
                column.add_spacing(6)
                column.add(path_row)
                column.add_spacing(6)
                column.add_stretch()
                self.content.add(column)

                self.__list_widget = list_widget

            def show(self):
                super().show()
                self.__list_widget.focused = True

        if len(items) == 0:
            # for initial startup (or with no preferences) try to use a default location
            # to avoid the user having to go through the horrendous choose dialog immediately.
            try:
                documents_dir = self.ui.get_document_location()
                workspace_dir = os.path.join(documents_dir, "Nion Swift Library")
                Cache.db_make_directory_if_needed(workspace_dir)
                path = os.path.join(workspace_dir, "Nion Swift Workspace.nslib")
                if not os.path.exists(path):
                    with open(path, "w") as fp:
                        json.dump({}, fp)
                if os.path.exists(path):
                    app.switch_library(workspace_dir)
                    return
            except Exception as e:
                pass

        choose_library_dialog = ChooseLibraryDialog(self.ui, self)
        choose_library_dialog.show()

    def create_document_controller(self, document_model, workspace_id, data_item=None):
        self._set_document_model(document_model)  # required to allow API to find document model
        document_controller = DocumentController.DocumentController(self.ui, document_model, workspace_id=workspace_id, app=self)
        self.__did_close_event_listeners[document_controller] = document_controller.did_close_event.listen(self.__document_controller_did_close)
        self.__create_new_event_listeners[document_controller] = document_controller.create_new_document_controller_event.listen(self.create_document_controller)
        self.__register_document_controller(document_controller)
        # attempt to set data item / group
        if data_item:
            display_panel = document_controller.selected_display_panel
            if display_panel:
                display_panel.set_display_panel_data_item(data_item)
        document_controller.show()
        return document_controller

    def __document_controller_did_close(self, document_controller):
        self.__did_close_event_listeners[document_controller].close()
        del self.__did_close_event_listeners[document_controller]
        self.__create_new_event_listeners[document_controller].close()
        del self.__create_new_event_listeners[document_controller]
        self.__document_controllers.remove(document_controller)

    def __register_document_controller(self, document_controller: DocumentController.DocumentController) -> None:
        assert document_controller not in self.__document_controllers
        self.__document_controllers.append(document_controller)
        # when a document window is registered, tell the menu handlers
        for menu_handler in self.__menu_handlers:  # use 'handler' to avoid name collision
            menu_handler(document_controller)

    @property
    def document_controllers(self) -> typing.List[DocumentController.DocumentController]:
        return copy.copy(self.__document_controllers)

    def register_menu_handler(self, new_menu_handler):
        assert new_menu_handler not in self.__menu_handlers
        self.__menu_handlers.append(new_menu_handler)
        # when a menu handler is registered, let it immediately know about existing menu handlers
        for document_controller in self.__document_controllers:
            new_menu_handler(document_controller)
        # return the menu handler so that it can be used to unregister (think: lambda)
        return new_menu_handler

    def unregister_menu_handler(self, menu_handler):
        self.__menu_handlers.remove(menu_handler)

    @property
    def menu_handlers(self) -> typing.List:
        return copy.copy(self.__menu_handlers)

    def run_all_tests(self):
        Test.run_all_tests()


def get_root_dir():
    root_dir = os.path.dirname((os.path.dirname(os.path.abspath(__file__))))
    path_ascend_count = 2
    for i in range(path_ascend_count):
        root_dir = os.path.dirname(root_dir)
    return root_dir
